import os
import shutil
import gzip
import tempfile
import typing as t
from typing import Optional, Dict, Tuple

import io
import requests.adapters
import yaml
import requests
from requests.auth import HTTPBasicAuth
from rich.progress import Progress
from typing import Union


class MockProgress:
    
    def add_task(self, *args, **kwargs):
        return 0
    
    def update(self, *args, **kwargs):
        pass
    
    def start_task(self, *args, **kwargs):
        pass
    
    
class AbstractFileShare:
    """
    This class defines the abstract interface for a file share server. This interface defines the methods 
    which can be used to interact with the remote file share server solution. The concrete implementation 
    and storage solution that is used for the file share server is not defined here and should be
    implemented in a subclass.
    """
    
    def __init__(self, url: str):
        self.url = url
        self.metadata: Optional[dict] = None
    
    def fetch_metadata(self) -> dict:
        """
        This method should fetch the metadata from the remote file share server and return the metadata 
        as a dictionary that was loaded from the metadata yml file. This method should also populate the 
        "self.metadata" attribute of the object.
        """
        raise NotImplementedError()

    def upload_metadata(self, metadata: dict) -> None:
        """
        This method should take the metadata dictionary and upload that information to the remote file 
        server such that after the function is executed the remote file share information should be 
        updated.
        """
        raise NotImplementedError()

    def download_dataset(self,
                         dataset_name: str,
                         folder_path: str,
                         ) -> None:
        """
        This method should download the dataset which is identified by the unique string name and the 
        downloaded dataset should be placed in the given folder path.
        """
        raise NotImplementedError()
    
    def authenticate(self, auth_info: dict) -> None:
        """
        This method should be implemented so that after calling it the object should be authenticated 
        in the file server - meaning that it should be possible to make modifications to the files there
        This is mostely relevant for developers and not regular users of the package.
        """
        raise NotImplementedError()
    
    # ~ implement dict-like behaviour
    # Here we implement that the file share server instance acts as a dictionary regarding the metadata
        
    def assert_metadata(self):
        """
        This method checks if the metadata attribute of the object has already been loaded and 
        raises an error if that is not the case.
        """
        if self.metadata is None:
            raise LookupError('Metadata has not been fetched from the remote location yet!'
                              'Call the "fetch_metadata" method on the file share object first!')
        
    def keys(self) -> t.List[str]:
        self.assert_metadata()
        return self.metadata.keys()
        
    def __contains__(self, key: str) -> bool:
        self.assert_metadata()
        return self.metadata[key]
        
    def __getitem__(self, key: str) -> t.Any:
        self.assert_metadata()
        return self.metadata[key]


class NextcloudFileShare(AbstractFileShare):
    """
    This specific subblass implements the usage of a Nextcloud server as a file share solution to 
    download the remote datasets from.
    
    :param url: The publically shared url from which the datasets can be downloaded.
    :param dav_url: The url to the nextcloud server DAV endpoint of the specific folder that contains 
        the shared datasets.
    :param dav_username: The username that is used to authenticate against the DAV endpoint of the
        nextcloud server.
    :param dav_password: The password that is used to authenticate against the DAV endpoint of the
        nextcloud server.
    :param use_download_url_path: If set to True, the files will be downloaded from the share point 
        using the URL directly instead of using the URL parameters to define the relative path 
        (legacy behaviour). This may be set to False for older versions of nextcloud.
    """
    def __init__(self, 
                 url: str,
                 dav_url: Optional[str] = None,
                 dav_username: Optional[str] = None,
                 dav_password: Optional[str] = None,
                 verify: bool = True,
                 use_download_url_path: bool = True,
                 **kwargs,
                 ) -> None:
        
        AbstractFileShare.__init__(self, url)
        
        self.dav_url = dav_url
        self.dav_username = dav_username
        self.dav_password = dav_password
        self.verify = verify
        self.use_download_url_path = use_download_url_path

        ## -- Derived Attributes --
        
        # The given URL will be in the format of a publically shared link to a folder on the nextcloud 
        # served in the format of "https://nextcloud.example.com/s/abc123xyz456"
        # We want to separate this into the base URL of the nextcloud server and the share token
        # Parse the given URL and reconstruct only the scheme and netloc (base URL)
        base_url, share_token = self.url.split('/s/', 1)
        self.base_url = base_url.rstrip('/')
        self.share_token = share_token.replace('/download', '').rstrip('/')

        ## -- Metadata-related Attributes --
        # This is the name of the file on the file share server which contains the metadata. This 
        # is a human-readable yml file which not only contains the information about the file share 
        # in general, but also detailed information about the individual datasets that are 
        # available for download.
        self.metadata_name = 'metadata.yml'
        # This is the path where the metadata file will be stored on the local system AFTER it has 
        # been downloaded from the file share server. So this file may or may not exist at the
        # time of the object creation, but it will definitely exist after the first call to the
        # fetch_metadata method.
        self.metadata_path: str = os.path.join(tempfile.gettempdir(), f'chemdata_{self.metadata_name}')
        
        # This is the dictionary where the content of the metadata file will be stored after it has
        # been downloaded from the file share server. This is a dictionary representation of the
        # metadata file. This dictionary will be created by the fetch_metadata method.
        self.metadata: Dict[str, t.Any] = None
    
    def fetch_metadata(self, force: bool = True) -> dict:
        """
        Fetches the "metadata.yml" file from the remote file server, returns the metadata dict and 
        also stores it in the metadata attribute of the object. If the metadata file already exists
        on the local system, it will be loaded from there instead of being downloaded again.
        
        :param force: If set to True, the metadata file will be downloaded again even if it already
        
        :returns: None
        """
        # For the sake of efficiency, we will first check if the metadata file already exists on the
        # local system. If it does, we will simply load it into the metadata attribute of the object
        # and return.
        if not os.path.exists(self.metadata_path) or force:
            # If the metadata file does not exist, we will download it from the file share server.
            content = self.download(self.metadata_name)
            with open(self.metadata_path, mode='wb') as file:
                file.write(content)
                
        # Now that the file definitely exists, we will load it into the metadata attribute of the object.
        with open(self.metadata_path, mode='r') as file:
            self.metadata = yaml.safe_load(file)
            
        return self.metadata
    
    def download_dataset(self, 
                         dataset_name: str, 
                         folder_path: str = tempfile.gettempdir(),
                         progress: Progress = MockProgress(),
                         ) -> str:
        """
        Given the string ``dataset_name`` of a dataset on the remote file share server, this method
        will download that dataset to the local system into the given ``folder_path``. Note that 
        the dataset name does NOT have to include the file extension!
        
        :param dataset_name: The string name of the dataset to be downloaded.
        :param folder_path: The string path to the folder where the downloaded dataset should be stored.
        :param progress: An optional Progress instance which will be used to track the download progress.
        
        :returns: The absolute path to the downloaded dataset on the local system.
        """
        
        if not dataset_name.endswith('.mpack'):
            dataset_name = dataset_name + '.mpack'
            
        file_name = dataset_name
            
        # 04.07.24
        # In the first instance we are going to try and download the compressed (gzip - gz) version 
        # of the dataset because that is usually at least 10x smaller and should therefore be a lot 
        # faster to download and only if that doesn't exist or fails due to some other issue we 
        # attempt to download the uncompressed version.
        try:
            file_name_compressed = f'{file_name}.gz'
            file_path_compressed = self.download_file(
                file_name_compressed,
                folder_path=folder_path,
                progress=progress,
            )
            
            # Then we can decompress the file using the gzip module. This may take a while.
            file_path = os.path.join(folder_path, file_name)
            with open(file_path, mode='wb') as file:
                with gzip.open(file_path_compressed, mode='rb') as compressed_file:
                    shutil.copyfileobj(compressed_file, file)
        
        # Otherwise we try to download the file without the compression
        except Exception:
            file_path = self.download_file(
                file_name, 
                folder_path=folder_path,
                progress=progress,
            )
            
        return file_path
            
    def download_file(self,
                      file_name: str,
                      folder_path: str = tempfile.gettempdir(),
                      progress: Progress = MockProgress(),
                      ) -> str:
        """
        Given the string ``file_name`` of a file on the remote file share server, this method 
        will download that file to the local system into the given ``folder_path``. Note that 
        the given file name has to include a file extension!
        
        :param file_name: The string name of the file to be downloaded.
        :param folder_path: The string path to the folder where the downloaded file should be stored.
        :param progress: An optional Progress instance which will be used to track the download 
            progress.
            
        :returns: The absolute path to the downloaded file on the local system.
        """
        file_path = os.path.join(folder_path, file_name)
        self.download(file_name, progress=progress, folder_path=folder_path)
            
        return file_path
    
    # ~ utility methods
    
    def construct_download_url(self,
                               file_name: str,
                               ) -> str:
        """
        Returns the direct download URL for a file stored on the file share server using the
        WebDAV public endpoint format.

        :param file_name: The name of the file to be downloaded, including any subdirectory structure
            and file extension, relative to the root of the shared folder.

        :returns: The fully constructed download URL as a string. This URL can be used in a browser or
            with command-line tools (such as wget or curl) to download the file directly from the server.
            The format is: ``{self.base_url}/public.php/dav/files/{self.share_token}/{file_name}``

        Example::

            share = NextcloudFileShare(url="https://nextcloud.example.com/s/abc123xyz456")
            url = share.construct_download_url("mydata.mpack.gz")
            # url == 'https://nextcloud.example.com/public.php/dav/files/abc123xyz456/mydata.mpack.gz'

        Note:
            This method uses the WebDAV public endpoint format introduced in Nextcloud 29, which is
            compatible with Nextcloud Hub 10 (Nextcloud 31) and newer versions. The old
            /s/{token}/download endpoint format is deprecated and no longer works reliably.
        """
        # Use the new WebDAV public endpoint format (available since Nextcloud 29)
        # This replaces the deprecated /s/{token}/download endpoint
        download_url = f'{self.base_url}/public.php/dav/files/{self.share_token}/{file_name}'

        return download_url
    
    def download(self, 
                 file_name: str, 
                 progress: Progress = MockProgress(),
                 folder_path: Union[str, None] = None,
                 ) -> bytes:
        """
        Given the full ``file_name`` of a file located in the current file share location, this 
        method will download that file. If the ``folder_path`` argument is given as a valid 
        folder location on the local system, the downloaded content will be written as a local
        file with the same filename into that folder (overwrites existing files). If the ``folder_path``
        is None, the bytes content of the file will instead be directly returned by this method.
        
        :param file_name: The string file name of the file to be downloaded - including the file 
            extension.
        :param progress: An optional rich.progress.Progress instance which will be used to track 
            the download progress in case it is given. If not explicticly provided, will be ignored.
        :param folder_path: An optional string path to a folder on the local system where the
            downloaded file should be stored. If not given, the downloaded content will be returned
            as a bytes object.
            
        :returns: The bytes content of the downloaded file if no folder path is given. Otherwise,
            returns None.
        """
        
        # This method will construct the download URL for the given file name 
        file_url: str = self.construct_download_url(file_name)
        
        # Create a session with no connection pooling
        session = requests.Session()
        session.keep_alive = False
        
        response = session.get(
            file_url, 
            stream=True, 
            headers={'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; rv:91.0) Gecko/20100101 Firefox/91.0', 'Connection': 'close'},
            verify=self.verify,
            timeout=5,
        )
        
        # Only when we get the correct status code, we can assume that the download was 
        # succesfull and we can proceed with the download.
        if response.status_code == 200:
            
            total_size = int(response.headers.get('Content-Length', 0))
            
            # Optionally, we update the progress of the download with the given Progress 
            # instance so that the progress can be tracked in the command line.
            task = progress.add_task(f'{file_name}', total=total_size)
            
            # 10.06.24
            # If a folder path is given we directly open a new file object in that folder
            # and store the downloaded information there. However, if there is no folder path 
            # given we want to return the downloaded content as a bytes object.
            
            # This distinction is new and we do this here because we want to consider 2 cases:
            # - For very large download sizes we really dont want to have to store the entire 
            #   file content into the memory of the system. Instead we want to write it directly
            #   to the disk to be more efficient.
            # - For very small download sizes there is the opposite argument: To be more efficient 
            #   we directly return the content as a bytes object instead of writing it to the disk
            #   first.
            if folder_path is None:
                file = io.BytesIO()
            else:
                file_path = os.path.join(folder_path, file_name)
                file = open(file_path, mode='wb')    
            
            with file:
                bytes_written = 0
                for chunk in response.iter_content(chunk_size=1024**2):
                    if chunk:
                        file.write(chunk)
                        progress.update(task, advance=bytes_written)
                        bytes_written += len(chunk)
                
                # If there is no file path given, we return the content of the file as a bytes 
                # object.
                if folder_path is None:
                    return file.getvalue()
        
        elif response.status_code == 404:
            raise FileNotFoundError(f'File not found on the server: {file_url}')
        
        else:
            raise Exception(f'Failed to download file from the server: {file_url}')

    def exists(self,
               file_name: str,
               ) -> Tuple[bool, dict]:
        """
        Checks whether a file with the given ``file_name`` exists on the remote Nextcloud file share server.

        This method sends a HEAD request to the Nextcloud WebDAV endpoint to determine if the specified file exists. If the file exists, it returns True and a dictionary containing basic metadata about the file (such as size, last modified date, and content type) as obtained from the response headers. If the file does not exist, it returns False and an empty dictionary.

        :param file_name: The name of the file to check for existence on the remote server. This should include any subdirectory structure and file extension, relative to the root of the shared folder.

        :returns: A tuple (exists, metadata) where:
            - exists (bool): True if the file exists on the server, False otherwise.
            - metadata (dict): If the file exists, a dictionary with the following keys:
                - 'size': The size of the file in bytes (as a string), or 'unknown' if not available.
                - 'last_modified': The last modified date of the file (as a string), or 'unknown' if not available.
                - 'content_type': The MIME type of the file (as a string), or 'unknown' if not available.
              If the file does not exist, this will be an empty dictionary.

        :raises AssertionError: If the required DAV credentials (dav_url, dav_username, dav_password) are not set on the object.
        :raises Exception: If the server returns an unexpected status code or there is a network error.

        Example::

            share = NextcloudFileShare(
                url="https://nextcloud.example.com/s/abc123xyz456",
                dav_url="https://nextcloud.example.com/remote.php/dav/files/username/shared_folder",
                dav_username="myuser",
                dav_password="mypassword"
            )
            exists, meta = share.exists("mydata.mpack.gz")
            if exists:
                print(f"File exists! Size: {meta['size']} bytes, Last modified: {meta['last_modified']}")
            else:
                print("File does not exist on the server.")
        """
        # This method will simply check if the file share object instance as it is has enough 
        # information / permission to actually upload files to the remote fileshare. 
        # Will raise an error if that is not the case.
        self.assert_dav(action='upload')
        
        ## -- Making Request --
        # The check for the existence of a file on a remote nextcloud file share server, one 
        # simply needs to make a HEAD request to the DAV endpoint.
        file_url = f'{self.dav_url}/{file_name}'
        response = requests.head(
            file_url,
            auth=HTTPBasicAuth(
                self.dav_username,
                self.dav_password,
            ),
            verify=self.verify,
            timeout=1.0,
        )
        
        ## -- Handling Response --
        # If the HEAD request was successful, that means that the file exists on the server and 
        # we can infer some basic metadata about the file from the response headers. 
        if response.status_code == 200:
            
            size = response.headers.get('Content-Length', 'unknown')
            last_modified = response.headers.get('Last-Modified', 'unknown')
            content_type = response.headers.get('Content-Type', 'unknown')
            
            return True, {
                'size': size,
                'last_modified': last_modified,
                'content_type': content_type,
            }
            
        else:
            
            return False, {}
        

    def upload(self,
               file_name: str,
               file_path: str,
               ) -> None:
        """
        Given the ``file_name`` that a file should have on the server and the ``file_path`` to the 
        local version of the file, this method uploads the file to the remote file share server, 
        using the DAV endpoint of the server.
        
        :param file_name: The string name of the file that the file should have on the server.
        :param file_path: The string path to the local file that should be uploaded
        
        :returns: None
        """
        # This method will simply check if the file share object instance as it is has enough 
        # information / permission to actually upload files to the remote fileshare. 
        # Will raise an error if that is not the case.
        self.assert_dav(action='upload')
        
        upload_url = f'{self.dav_url}/{file_name}'
        with open(file_path, 'rb') as file:
            
            response = requests.put(
                upload_url,
                data=file,
                auth=HTTPBasicAuth(
                    self.dav_username,
                    self.dav_password,
                ),
                verify=self.verify,
            )
            
            # If something went wrong during the upload, we raise an error.
            response.raise_for_status()

    def assert_dav(self, action: str = '') -> None:
        
        assert self.dav_url is not None, (
            f'The requested action "{action}" requires extended access to the nextcloud server and therefore requires the '
            f'dav_url to be given to the NextcloudFileShare object. Please provide the dav_url parameter to the constructor '
            f'or by adding it to the config file.'
        )
        
        assert self.dav_username is not None, (
            f'The requested action "{action}" requires extended access to the nextcloud server and therefore requires the '
            f'dav_username to be given to the NextcloudFileShare object. Please provide the dav_username parameter to the '
            f'constructor or by adding it to the config file.'
        )
        
        assert self.dav_password is not None, (
            f'The requested action "{action}" requires extended access to the nextcloud server and therefore requires the '
            f'dav_password to be given to the NextcloudFileShare object. Please provide the dav_password parameter to the '
            f'constructor or by adding it to the config file.'
        )
        

# This dictionary assigns the string names of the different file share types to the actual classes 
# So given the name of a fileshare type, this map can be used to obtain to actual class to then 
# consruct a fileshare instance from.
FILESHARE_TYPES: Dict[str, type] = {
    'nextcloud': NextcloudFileShare,
}        



def construct_file_share(file_share_type: str,
                         file_share_url: str,
                         file_share_kwargs: dict,
                         ) -> AbstractFileShare:
    """
    Given the string name of a valid ``file_share_type`` and the ``file_share_url`` of the file share
    server, this function will construct an instance of the corresponding file share class which is 
    a subclass implementing the AbstractFileShare interface.
    
    :param file_share_type: The unique string name that identifies the type of file share server that 
        is to be used.
    :param file_share_url: The url that points to the actual file share server. This is a required 
        parameter for the construction of any file share.
    :param file_share_kwargs: A dictionary of additional keyword arguments that may be required for the
        construction of a specific file share subclass.
    """
    fileshare_class: type = FILESHARE_TYPES[file_share_type]
    fileshare: AbstractFileShare = fileshare_class(
        url=file_share_url,
        **file_share_kwargs,
    )
    return fileshare



"""
Unit tests for the "connectors.py" module that contains online data source connectors.
"""
import os
import tempfile
import zipfile
import pytest
from unittest.mock import Mock, patch, mock_open
from io import BytesIO

from chem_mat_data.connectors import ZenodoSource


class TestZenodoSource:
    """
    Test suite for the ZenodoSource class that handles downloading and extracting
    files from Zenodo repositories.
    """

    def test_init(self):
        """
        Test that ZenodoSource initializes correctly with url and relative_path.
        """
        url = "https://zenodo.org/record/123456/files/dataset.zip"
        relative_path = "data/molecules.csv"

        source = ZenodoSource(url, relative_path)

        assert source.url == url
        assert source.relative_path == relative_path
        assert source.temp_dir is None

    def test_prepare_creates_temp_dir(self):
        """
        Test that prepare() creates a temporary directory.
        """
        source = ZenodoSource("https://test.com/file.zip", "test.csv")

        source.prepare()

        assert source.temp_dir is not None
        assert os.path.exists(source.temp_dir)
        assert source.temp_dir.startswith(tempfile.gettempdir())

        # Cleanup
        source.cleanup()

    @patch('requests.get')
    def test_fetch_successful_download(self, mock_get):
        """
        Test successful download and extraction of a file from a mocked ZIP archive.
        """
        # Create a mock ZIP file content
        zip_buffer = BytesIO()
        with zipfile.ZipFile(zip_buffer, 'w') as zip_file:
            zip_file.writestr('data/molecules.csv', 'smiles,property\nCCO,1.2\nCCC,2.1\n')
        zip_content = zip_buffer.getvalue()

        # Mock the response
        mock_response = Mock()
        mock_response.iter_content.return_value = [zip_content]
        mock_response.raise_for_status.return_value = None
        mock_get.return_value = mock_response

        source = ZenodoSource(
            "https://zenodo.org/record/123456/files/dataset.zip",
            "data/molecules.csv"
        )

        file_path = source.fetch()

        # Verify the file was extracted correctly
        assert os.path.exists(file_path)
        with open(file_path, 'r') as f:
            content = f.read()
        assert 'smiles,property' in content
        assert 'CCO,1.2' in content

        # Verify requests.get was called correctly
        mock_get.assert_called_once_with(source.url, stream=True)

        # Cleanup
        source.cleanup()

    @patch('requests.get')
    def test_fetch_file_not_found_in_archive(self, mock_get):
        """
        Test that ValueError is raised when the specified file is not found in the archive.
        """
        # Create a mock ZIP file without the target file
        zip_buffer = BytesIO()
        with zipfile.ZipFile(zip_buffer, 'w') as zip_file:
            zip_file.writestr('other/file.txt', 'some content')
        zip_content = zip_buffer.getvalue()

        mock_response = Mock()
        mock_response.iter_content.return_value = [zip_content]
        mock_response.raise_for_status.return_value = None
        mock_get.return_value = mock_response

        source = ZenodoSource(
            "https://zenodo.org/record/123456/files/dataset.zip",
            "data/molecules.csv"
        )

        with pytest.raises(ValueError, match="File 'data/molecules.csv' not found in the downloaded archive"):
            source.fetch()

        source.cleanup()

    @patch('requests.get')
    def test_fetch_finds_file_by_basename(self, mock_get):
        """
        Test that fetch can find a file by its basename when the exact path doesn't match.
        """
        # Create a mock ZIP file with file in different location
        zip_buffer = BytesIO()
        with zipfile.ZipFile(zip_buffer, 'w') as zip_file:
            zip_file.writestr('different/path/molecules.csv', 'smiles,property\nCCO,1.2\n')
        zip_content = zip_buffer.getvalue()

        mock_response = Mock()
        mock_response.iter_content.return_value = [zip_content]
        mock_response.raise_for_status.return_value = None
        mock_get.return_value = mock_response

        source = ZenodoSource(
            "https://zenodo.org/record/123456/files/dataset.zip",
            "data/molecules.csv"  # This exact path doesn't exist, but basename does
        )

        file_path = source.fetch()

        assert os.path.exists(file_path)
        with open(file_path, 'r') as f:
            content = f.read()
        assert 'smiles,property' in content

        source.cleanup()

    @patch('requests.get')
    def test_fetch_handles_http_error(self, mock_get):
        """
        Test that HTTP errors are properly raised during download.
        """
        mock_response = Mock()
        mock_response.raise_for_status.side_effect = Exception("HTTP 404 Not Found")
        mock_get.return_value = mock_response

        source = ZenodoSource(
            "https://zenodo.org/record/nonexistent/files/dataset.zip",
            "data/molecules.csv"
        )

        with pytest.raises(Exception, match="HTTP 404 Not Found"):
            source.fetch()

        source.cleanup()

    def test_context_manager(self):
        """
        Test that ZenodoSource works correctly as a context manager.
        """
        url = "https://zenodo.org/record/123456/files/dataset.zip"
        relative_path = "data/molecules.csv"

        with ZenodoSource(url, relative_path) as source:
            assert source.url == url
            assert source.relative_path == relative_path
            # Context manager should work without calling prepare/cleanup manually

    @patch('requests.get')
    def test_context_manager_with_fetch(self, mock_get):
        """
        Test context manager with actual fetch operation and automatic cleanup.
        """
        # Create a mock ZIP file content
        zip_buffer = BytesIO()
        with zipfile.ZipFile(zip_buffer, 'w') as zip_file:
            zip_file.writestr('test.txt', 'test content')
        zip_content = zip_buffer.getvalue()

        mock_response = Mock()
        mock_response.iter_content.return_value = [zip_content]
        mock_response.raise_for_status.return_value = None
        mock_get.return_value = mock_response

        temp_dir_path = None

        with ZenodoSource(
            "https://zenodo.org/record/123456/files/dataset.zip",
            "test.txt"
        ) as source:
            file_path = source.fetch()
            temp_dir_path = source.temp_dir

            assert os.path.exists(file_path)
            assert os.path.exists(temp_dir_path)

        # After exiting context, temp directory should be cleaned up
        assert not os.path.exists(temp_dir_path)

    def test_cleanup_removes_temp_dir(self):
        """
        Test that cleanup() properly removes the temporary directory.
        """
        source = ZenodoSource("https://test.com/file.zip", "test.csv")
        source.prepare()

        temp_dir = source.temp_dir
        assert os.path.exists(temp_dir)

        source.cleanup()

        assert not os.path.exists(temp_dir)
        assert source.temp_dir is None

    def test_cleanup_handles_nonexistent_dir(self):
        """
        Test that cleanup() handles case where temp_dir doesn't exist.
        """
        source = ZenodoSource("https://test.com/file.zip", "test.csv")
        source.temp_dir = "/nonexistent/path"

        # Should not raise an exception
        source.cleanup()
        assert source.temp_dir is None

    def test_multiple_cleanup_calls(self):
        """
        Test that multiple cleanup() calls don't cause issues.
        """
        source = ZenodoSource("https://test.com/file.zip", "test.csv")
        source.prepare()

        source.cleanup()
        source.cleanup()  # Second call should be safe

        assert source.temp_dir is None


class TestZenodoSourceIntegration:
    """
    Integration tests that use real network calls to Zenodo.
    These tests are marked to be skipped during normal test runs.
    """

    @pytest.mark.slow
    @pytest.mark.skipif(
        os.environ.get("SKIP_INTEGRATION_TESTS", "true").lower() == "true",
        reason="Integration tests skipped (set SKIP_INTEGRATION_TESTS=false to enable)"
    )
    def test_real_zenodo_download(self):
        """
        Test downloading from a real Zenodo repository.

        Uses: https://zenodo.org/records/11580890/files/TheJacksonLab/ClosedLoopTransfer-ClosedLoopTransfer.zip
        This test is marked as slow and skipped by default.
        """
        url = "https://zenodo.org/records/11580890/files/TheJacksonLab/ClosedLoopTransfer-ClosedLoopTransfer.zip?download=1"

        with ZenodoSource(url, "README.md") as source:
            try:
                file_path = source.fetch()

                # Verify the file was downloaded and extracted
                assert os.path.exists(file_path)
                assert file_path.endswith("README.md")

                # Check that file has some content
                with open(file_path, 'r') as f:
                    content = f.read()
                assert len(content) > 0

            except Exception as e:
                pytest.skip(f"Real Zenodo download failed (network issue?): {e}")

    @pytest.mark.slow
    @pytest.mark.skipif(
        os.environ.get("SKIP_INTEGRATION_TESTS", "true").lower() == "true",
        reason="Integration tests skipped (set SKIP_INTEGRATION_TESTS=false to enable)"
    )
    def test_real_zenodo_file_search(self):
        """
        Test the file search functionality with a real Zenodo archive.
        This tests the fallback mechanisms for finding files when exact paths don't match.
        """
        url = "https://zenodo.org/records/11580890/files/TheJacksonLab/ClosedLoopTransfer-ClosedLoopTransfer.zip?download=1"

        with ZenodoSource(url, "LICENSE") as source:
            try:
                file_path = source.fetch()

                assert os.path.exists(file_path)

                # Verify it found a LICENSE file
                with open(file_path, 'r') as f:
                    content = f.read()
                assert len(content) > 0
                # LICENSE files typically contain copyright information
                assert any(word in content.lower() for word in ['license', 'copyright', 'mit', 'apache', 'gpl'])

            except Exception as e:
                pytest.skip(f"Real Zenodo download failed (network issue?): {e}")


import os
from typing import Optional, Dict, Any

import tomlkit
from dotenv import load_dotenv
from appdirs import AppDirs

from chem_mat_data.cache import Cache
from chem_mat_data.utils import config_file_from_template


class Singleton(type):
    """
    This is metaclass definition, which implements the singleton pattern. The objective is that whatever
    class uses this as a metaclass does not work like a traditional class anymore, where upon calling the
    constructor a NEW instance is returned. This class overwrites the constructor behavior to return the
    same instance upon calling the constructor. This makes sure that always just a single instance
    exists in the runtime!

    **USAGE**
    To implement a class as a singleton it simply has to use this class as the metaclass.
    .. code-block:: python
        class MySingleton(metaclass=Singleton):
            def __init__(self):
                # The constructor still works the same, after all it needs to be called ONCE to create the
                # the first and only instance.
                pass
        # All of those actually return the same instance!
        a = MySingleton()
        b = MySingleton()
        c = MySingleton()
        print(a is b) # true
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
            
        return cls._instances[cls]


class Config(metaclass=Singleton):
    
    def __init__(self):
        
        # ~ loading environment variables
        
        # By using the override=False option the environment variable definitions in the .env file are 
        # only used if definitions for those variables do not already exist in the environment! This 
        # is useful to prevent overwriting of environment variables if the user wants to override some of 
        # them with custom values.
        load_dotenv(override=False)
        self.fileshare_url = os.getenv('FILESHARE_URL')
        
        # This object can be used to access the absolute paths of application specific directories like 
        # the config folder or a cache folder in an OS agnostic way.
        self.app_dirs = AppDirs('chem_mat_data')
        
        # ~ local dataset caching
        # The cache can be used to store datasets that have already been downloaded so that they don't 
        # have to be downloaded every time. This path attribute will contain the absolute string path
        # to the cache folder.
        self.cache_path: Optional[str] = None
        # The Cache object instance is a wrapper that provides a more convenient way to interact with the
        # cache folder. It provides methods to add datasets to the cache, check if a dataset is already in
        # the cache, and to clear the cache, etc.
        self.cache: Optional[Cache] = None
        # This method will change the cache path to the given path, which also includes the construction
        # of a new Cache object instance to manage the access to that cache folder.
        self.set_cache_path(self.app_dirs.user_cache_dir)
        
        # ~ config files
        self.config_path: Optional[str] = None
        self.config_file_path: Optional[str] = None
        self.config_data: Optional[Dict[str, Any]] = None
        self.set_config_path(self.app_dirs.user_config_dir)
        
    def set_cache_path(self, cache_path: str) -> None:
        """
        This method will change the cache path to the given ``cache_path``, which also includes the construction
        of a new Cache object instance to manage the access to that cache folder.
        """
        self.cache_path = cache_path
        if not os.path.exists(self.cache_path):
            os.makedirs(self.cache_path)
        
        self.cache =  Cache(self.cache_path)

    def set_config_path(self, config_path: str) -> None:
        
        self.config_path = config_path
        if not os.path.exists(self.config_path):
            os.makedirs(self.config_path)
            
        # Inside this config folder there needs to be a "config.yml" file which contains the configuration
        # data for the application.
        self.config_file_path = os.path.join(self.config_path, 'config.toml')
        if not os.path.exists(self.config_file_path):
            # This util function will create a new default config file at the given file path
            config_file_from_template(self.config_file_path)
                
        # Finally, we can load the configuration data from the config file and load the results into the 
        # "config_data" dict type attribute.
        with open(self.config_file_path) as file:
            self.config_data = tomlkit.loads(file.read())
            
    def save(self) -> None:
        """
        This method will save the current content of the "config_data" attribute to the config TOML file 
        on the disk.
        
        :returns: None
        """
        with open(self.config_file_path, 'w') as file:
            file.write(tomlkit.dumps(self.config_data))
    
    def load(self):
        """
        This method will load the content of the config TOML file into the "config_data" attribute.
        
        :returns: None
        """
        with open(self.config_file_path) as file:
            self.config_data = tomlkit.loads(file.read())
    
    # ~ actual config values
    # The following methods define the interface to access the actual configuration values defined in the config 
    # file.

    def get_fileshare_url(self) -> str:
        """
        Returns the fileshare URL which was loaded from the environment variables. This URL points to 
        a cloud folder which contains all the actual dataset files and from where the datasets will
        be downloaded ultimately.
        
        :returns: string absolute URL
        """
        return self.config_data['remote']['fileshare_url']
    
    def get_fileshare_type(self) -> str:
        """
        Returns the fileshare type which was loaded from the config file. This type specifies the
        type of fileshare that should be used to download the datasets (e.g. Nextcloud, Zenodo, etc...)
        """
        return self.config_data['remote']['fileshare_type']
    
    def get_fileshare_parameters(self, fileshare_type: str) -> dict:
        """
        Returns a dict with the additional parameters specifically for the given "fileshare_type".
        
        :returns: dict
        """
        return self.config_data['remote'].get(fileshare_type, {})


import os
import sys
import datetime
import difflib
import gzip
import shutil
import tempfile
import typing as t
from typing import Any, List, Tuple, Dict, Union, Optional
from collections import Counter

import numpy as np
import pandas as pd
from rdkit import Chem
from rdkit.Chem import Descriptors
from rdkit.Chem.Scaffolds import MurckoScaffold

import rich
import rich_click as click
from rich import box
from rich.console import Console, ConsoleOptions, Group
from rich.panel import Panel
from rich.style import Style
from rich.padding import Padding
from rich.text import Text
from rich.syntax import Syntax
from rich.table import Table
from rich.progress import Progress
from rich.rule import Rule
from rich.columns import Columns

from chem_mat_data.utils import get_version
from chem_mat_data.utils import check_version_compatible
from chem_mat_data.utils import TEMPLATE_PATH
from chem_mat_data.utils import RichMixin
from chem_mat_data.utils import open_file_in_editor
from chem_mat_data.config import Config
from chem_mat_data.web import NextcloudFileShare

from dotenv import load_dotenv


# Load environment variables from a .env file if it exits and get the server URL from environment variables
load_dotenv()
server_url = os.getenv("url")
if not server_url:
    raise ValueError("Server URL not found in environment variables")


# == RICH WIDGETS ==
# Reusable small display components used across multiple views.


class LevelIndicator:
    """
    A small bar widget that visualizes a numeric value as filled/empty squares
    against a set of thresholds.

    Renders as e.g. ``■■■□□`` where filled squares are colored with ``color``
    and empty squares are dim grey. The number of levels equals
    ``len(thresholds) + 1``.

    :param value: The numeric value to visualize.
    :param thresholds: Sorted list of boundary values between levels.
    :param color: Rich color name for filled squares.
    """

    def __init__(self, value: int, thresholds: list, color: str = 'green'):
        self.value = value
        self.thresholds = thresholds
        self.color = color

    @property
    def level(self) -> int:
        return sum(1 for t in self.thresholds if self.value >= t) + 1

    @property
    def total(self) -> int:
        return len(self.thresholds) + 1

    def __rich__(self) -> Text:
        filled = Text('■' * self.level, style=self.color)
        empty = Text('□' * (self.total - self.level), style='grey30')
        result = Text()
        result.append_text(filled)
        result.append_text(empty)
        return result


#: Default thresholds for dataset compound counts.
COMPOUND_THRESHOLDS = [1_000, 10_000, 50_000, 500_000]


# == RICH DISPLAY ELEMENTS ==
# The following classes are used to define the rich display elements that are used to render the
# different parts of the command line interface. These classes are all subclasses of the RichMixin
# class which is a simple class that defines the __rich_console__ method which is used to render
# the object to the console using the rich library.


class RichLogo:
    """
    A rich display which will show the ASlurmX logo in ASCII art when printed.
    """

    STYLE = Style(bold=True, color="white")

    def __rich_console__(self, console, options):
        text_path = os.path.join(TEMPLATE_PATH, "logo_text.txt")
        with open(text_path) as file:
            text_string: str = file.read()
            text = Text(text_string, style=self.STYLE)
            
        image_path = os.path.join(TEMPLATE_PATH, "logo_image.txt")
        with open(image_path) as file:
            image_string: str = file.read()
            # Replace \e with actual escape character and create Text from ANSI
            ansi_string = image_string.replace('\\e', '\033')
            image = Text.from_ansi(ansi_string)
            
        side_by_side = Columns([image, text], equal=True, padding=(0, 3))
        yield Padding(side_by_side, (1, 3, 0, 3))
            
            
class RichHelp(RichMixin):
    """
    Implements the "rich" console rendering of the help text for the command line interface. This help 
    section consists of a brief general description of the CLI and also a brief section containing 
    example use cases of the commands.
    """
    
    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:
        yield 'Command Line Interface to the ChemMatDatabase.\n'
        yield Text((
            'The database consists of a collection of datasets chemistry and materials sience. Each datasets contains '
            'various molecules and/or crystal structures associated with a specific property. The main purpose of these '
            'datasets is to be used for the training of machine learning models for the prediction of these properties.\n'
        ))
        yield Text((
            'You can print a list of the available datasets using the "list" command.'
        ), style=Style(color='bright_black'))
        yield Padding(Syntax((
            'cmdata list'
        ), lexer='bash', theme='monokai', line_numbers=False), (1, 5))
        yield Text((
            'To download a dataset, use the "download" command followed by the name of the dataset.'
        ), style=Style(color='bright_black'))
        yield(Padding(Syntax((
            'cmdata download "dataset_name"'
        ), lexer='bash', theme='monokai', line_numbers=False), (1, 5)))
        
        
class RichCommands(RichMixin):

    def __init__(self, commands: Dict[str, Union[click.RichCommand, click.RichGroup]]):
        self.commands = commands

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:

        # Separate commands and groups
        commands = {}
        groups = {}

        for command_name, command in self.commands.items():
            # A command group has subcommands, while a regular command doesn't
            if hasattr(command, 'commands') and command.commands:
                groups[command_name] = command
            else:
                # Regular commands don't have subcommands
                commands[command_name] = command

        # Display regular commands in one panel
        if commands:
            table_commands = Table(
                title=None,
                box=None,
                show_header=False,
                padding=(0, 1),
                expand=True,
            )
            # Force fixed width by setting both min and max to the same value
            table_commands.add_column("Command", style="bold cyan", min_width=20, max_width=20, no_wrap=True)
            table_commands.add_column("Description", style="white", ratio=1)

            for command_name, command in commands.items():
                table_commands.add_row(
                    command_name,
                    command.short_help or ''
                )

            panel_commands = Panel(
                table_commands,
                title='[bold]Commands[/bold]',
                title_align='left',
                style='bright_black'
            )
            yield panel_commands

        # Display each command group in its own panel with subcommands
        for group_name, group in groups.items():
            # Create a table for this command group with fixed width for command column
            table = Table(
                title=None,
                box=None,
                show_header=False,
                padding=(0, 1),
                expand=True,
            )
            # Force fixed width by setting both min and max to the same value
            table.add_column("Command", style="bold cyan", min_width=20, max_width=20, no_wrap=True)
            table.add_column("Description", style="white", ratio=1)

            # Add all subcommands of this group to the table
            for cmd_name, cmd in group.commands.items():
                # Try to get help text from various sources
                short_help = cmd.short_help
                if not short_help and cmd.help:
                    # Extract first meaningful line from the full help text
                    help_lines = [line.strip() for line in cmd.help.split('\n') if line.strip()]
                    short_help = help_lines[0] if help_lines else ''
                if not short_help:
                    short_help = ''

                # Show composite command (e.g., "cache list" instead of just "list")
                composite_cmd = f"{group_name} {cmd_name}"
                table.add_row(composite_cmd, short_help)

            # Wrap the table in a panel with the group name as title
            panel = Panel(
                table,
                title=f'[bold]{group_name.capitalize()}[/bold]',
                title_align='left',
                style='bright_black',
            )

            yield panel
        

class RichDatasetInfo(RichMixin):
    """
    Implements the "rich" console rendering of the specific information about a single dataset.

    Uses a modern card-style layout with a colored hero panel whose appearance is driven
    by the dataset category, followed by lightweight rule-delimited sections for
    description, references, and tags.
    """

    # Maps dataset categories to their accent color used for borders, badges, etc.
    CATEGORY_COLORS: dict = {
        'organic':   'green',
        'tmc':       'magenta',
    }
    DEFAULT_COLOR: str = 'cyan'

    def __init__(self, name: str, info: dict):
        self.name = name
        self.info = info

    def _color(self) -> str:
        """Return the accent color for the current dataset's category."""
        category = self.info.get('category', 'organic')
        return self.CATEGORY_COLORS.get(category, self.DEFAULT_COLOR)

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:

        color = self._color()
        category = self.info.get('category', 'organic')

        yield ''

        # -- Hero panel -------------------------------------------------------
        # Title (left): dataset name as a colored badge (background color, black text)
        # Subtitle (right): category in accent color text

        # Build the stats line: "9,889 ■■□□□ compounds  ·  1 target"
        compounds = self.info.get('compounds', 0)
        indicator = LevelIndicator(compounds, COMPOUND_THRESHOLDS, color=color)

        stats_line = Text.from_markup(f'[bold]{compounds:,}[/bold] ')
        stats_line.append_text(indicator.__rich__())
        stats_line.append(' compounds')

        targets = self.info.get('targets')
        if targets is not None:
            stats_line.append_text(Text.from_markup(
                f'  ·  [bold]{targets}[/bold] {"target" if str(targets) == "1" else "targets"}'
            ))

        # Type selector widget: shows all possible types with the active one(s) highlighted
        all_types = ['regression', 'classification', 'bioactivity']
        active_types = set(self.info.get('target_type', []))
        # Infer bioactivity from tags
        tags_lower = {t.lower() for t in self.info.get('tags', [])}
        if 'bioactivity' in tags_lower or 'virtual screening' in tags_lower:
            active_types.add('bioactivity')

        type_parts = Text()
        type_parts.append('Task ', style='dim')
        for i, t in enumerate(all_types):
            if i > 0:
                type_parts.append(' ')
            if t in active_types:
                type_parts.append_text(Text.from_markup(f'[{color} on grey23] {t} [/{color} on grey23]'))
            else:
                type_parts.append_text(Text.from_markup(f'[dim on grey15] {t} [/dim on grey15]'))

        # Format selector widget: shows all possible formats with available ones highlighted
        all_formats = ['csv', 'xyz', 'tmc', 'graph']
        raw_formats = set(self.info.get('raw', []))
        has_graph = self.info.get('full', False)
        active_formats = set()
        for fmt in raw_formats:
            if fmt in ('xyz', 'xyz_bundle'):
                active_formats.add('xyz')
            else:
                active_formats.add(fmt)
        if has_graph:
            active_formats.add('graph')
        if category == 'tmc':
            active_formats.add('tmc')

        format_parts = Text()
        format_parts.append('Data ', style='dim')
        for i, fmt in enumerate(all_formats):
            if i > 0:
                format_parts.append(' ')
            label = fmt.upper()
            if fmt in active_formats:
                format_parts.append_text(Text.from_markup(f'[{color} on grey23] {label} [/{color} on grey23]'))
            else:
                format_parts.append_text(Text.from_markup(f'[dim on grey15] {label} [/dim on grey15]'))

        # Version compatibility warning (if applicable)
        min_ver = self.info.get('min_version')
        version_line = None
        if min_ver:
            compatible = check_version_compatible(min_ver)
            if not compatible:
                version_line = Text.from_markup(
                    f'[bold red]requires >= {min_ver}  (installed: {get_version()})[/bold red]'
                )

        # Compose hero content — use a table for the selector rows to get
        # tighter vertical spacing (leading=0, no extra blank lines).
        selector_table = Table(box=None, show_header=False, padding=(0, 0, 1, 0), pad_edge=False)
        selector_table.add_column()
        selector_table.add_row(type_parts)
        selector_table.add_row(format_parts)

        verbose = self.info.get('verbose', '')
        hero_lines = []
        if verbose:
            hero_lines.append(Text(verbose, style='dim italic'))
        hero_lines.extend([stats_line, Text(''), selector_table])
        if version_line:
            hero_lines.append(Text(''))
            hero_lines.append(version_line)

        hero = Panel(
            Group(*hero_lines),
            title=f'[bold black on {color}] {self.name} [/bold black on {color}] [{color}]──[/{color}] [{color}]{category}[/{color}]',
            title_align='left',
            border_style=color,
            padding=(1, 1, 1, 2),
        )
        yield hero

        # -- Description ------------------------------------------------------
        yield Padding(
            Text(self.info['description'], style='white'),
            (1, 3),
        )

        # -- Notes (optional) -------------------------------------------------
        if self.info.get('notes'):
            yield Padding(
                Rule(title='Notes', style='dim', align='left'),
                (1, 1),
            )
            for note in self.info['notes']:
                yield Padding(
                    Text.from_markup(f'[dim]•[/dim]  {note}'),
                    (0, 3),
                )

        # -- Target descriptions (optional) -----------------------------------
        if self.info.get('target_descriptions'):
            yield Padding(
                Rule(title='Targets', style='dim', align='left'),
                (1, 1),
            )
            target_table = Table(
                box=None, show_header=False, padding=(0, 2), pad_edge=False,
            )
            target_table.add_column(justify='right', style=f'bold {color}', no_wrap=True)
            target_table.add_column(justify='left', style='white', no_wrap=False)

            for idx in sorted(self.info['target_descriptions'].keys(), key=int):
                target_table.add_row(str(idx), self.info['target_descriptions'][idx])

            yield Padding(target_table, (0, 3))

        # -- References (optional) --------------------------------------------
        if self.info.get('sources'):
            yield Padding(
                Rule(title='References', style='dim', align='left'),
                (1, 1),
            )
            for source in self.info['sources']:
                yield Padding(
                    Text.from_markup(f'[yellow]▸  {source}[/yellow]'),
                    (0, 3),
                )

        # -- Tags -------------------------------------------------------------
        if self.info.get('tags'):
            yield Padding(
                Rule(title='Tags', style='dim', align='left'),
                (1, 1),
            )
            tag_texts = [
                f'[on grey23] :label:  {tag} [/on grey23]'
                for tag in self.info['tags']
            ]
            yield Padding(
                Text.from_markup('  '.join(tag_texts)),
                (0, 3),
            )

        yield ''
        
        
class RichDatasetList(RichMixin):
    """
    Implements the "rich" console rendering of the specific information about a single dataset.


    This display element will show a table which contains one row for each dataset and columns 
    that contain various pieces of information about the datasets.
    """
    def __init__(self, 
                 datasets: Dict[str, Dict],
                 sort: bool = False,
                 show_hidden: bool = False,
                 ):
        self.datasets = datasets
        self.sort = sort
        self.show_hidden = show_hidden
        
    # Category color mapping (shared with RichDatasetInfo)
    CATEGORY_COLORS: dict = {
        'organic':   'green',
        'tmc':       'magenta',
    }
    DEFAULT_COLOR: str = 'cyan'

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:

        # Count visible datasets for the title
        visible = [
            n for n in self.datasets
            if self.show_hidden or not n.startswith('_')
        ]
        count = len(visible)

        yield ''
        yield Rule(
            title=f'[bold black on white] Available Datasets ({count}) [/bold black on white]',
            style='bold white',
            align='center',
        )

        table = Table(expand=True, box=box.SIMPLE_HEAVY, padding=(0, 1), show_edge=False)

        table.add_column("Name", justify="left", no_wrap=True, min_width=12)
        table.add_column("Cat.", justify="left", no_wrap=True, max_width=10)
        table.add_column("Description", justify="left", style="white", no_wrap=True, ratio=1)
        table.add_column("Compounds", justify="right", style="white", no_wrap=True)
        table.add_column("Targets", justify="right", style="white", no_wrap=True)
        table.add_column("Type", justify="center", no_wrap=True)

        names = list(self.datasets.keys())
        if self.sort:
            names.sort(key=lambda value: value.lower())

        for name in names:
            if not self.show_hidden and name.startswith('_'):
                continue

            details = self.datasets[name]
            category = details.get('category', 'organic')
            color = self.CATEGORY_COLORS.get(category, self.DEFAULT_COLOR)

            # Check version compatibility
            min_ver = details.get('min_version')
            compatible = True
            if min_ver:
                compatible = check_version_compatible(min_ver)

            if compatible:
                name_display = f'[bold {color}]{name}[/bold {color}]'
            else:
                name_display = f'[dim]{name} (>={min_ver})[/dim]'

            category_display = f'[{color}]{category}[/{color}]'

            # Abbreviated type badges
            type_badges = []
            for tt in details.get('target_type', []):
                if tt == 'regression':
                    type_badges.append(f'[yellow on grey15] REG  [/yellow on grey15]')
                elif tt == 'classification':
                    type_badges.append(f'[blue on grey15] CLS  [/blue on grey15]')
                else:
                    type_badges.append(f'[white on grey15] {tt[:3].upper()}  [/white on grey15]')

            compounds = details.get('compounds', 0)
            targets = details.get('targets', 'N/A')

            # Size indicator: 5 levels based on compound count thresholds
            indicator = LevelIndicator(compounds, COMPOUND_THRESHOLDS, color=color)
            size_text = Text(f'{compounds:,}  ' if isinstance(compounds, int) else str(compounds))
            size_text.append_text(indicator.__rich__())

            row_style = None if compatible else 'dim'
            table.add_row(
                name_display,
                category_display,
                details.get('verbose', '-'),
                size_text,
                str(targets),
                ' '.join(type_badges),
                style=row_style,
            )

        yield table
        yield Rule(style='bold')


class RichCacheList(RichMixin):

    """
    Implements the "rich" console rendering of all the lists that are stored in the cache.

    This display element will show a table which contains one row for each dataset and columns
    that contain various pieces of information about the datasets.
    """
    def __init__(self,
                 dataset_metadata_map: Dict[Tuple[str, str], Dict],
                 dataset_sizes: Dict[Tuple[str, str], int],
                 total_cache_size: int,
                 sort: bool = True,
                 show_hidden: bool = False,
                 ):
        self.dataset_metadata_map = dataset_metadata_map
        self.dataset_sizes = dataset_sizes
        self.total_cache_size = total_cache_size
        self.sort = sort
        self.show_hidden = show_hidden
        
    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:

        table = Table(title='Dataset Cache', expand=True, box=box.HORIZONTALS)

        table.add_column("Dataset", justify="left", style="cyan", no_wrap=True)
        table.add_column("Created", justify="left", style="bright_black")
        table.add_column("Size (MB)", justify="right", style="magenta")

        keys = list(self.dataset_metadata_map.keys())
        # potentially we want to sort the items depending on the "sort" argument
        if self.sort:
            keys.sort()

        for key in keys:

            info: dict = self.dataset_metadata_map[key]
            name, typ = key

            dt = datetime.datetime.fromtimestamp(info['_cache_time'])

            # Get dataset size in MB
            size_bytes = self.dataset_sizes.get(key, 0)
            size_mb = size_bytes / (1024 * 1024)

            table.add_row(
                f'{name}.{typ}',
                f'{dt:%d.%b %Y, %H:%M}',
                f'{size_mb:.2f}',
            )

        yield table

        # Add total cache size information
        total_mb = self.total_cache_size / (1024 * 1024)
        yield ''
        yield Text(f'Total cache size: {total_mb:.2f} MB', style='bold green')


class RichCacheInfo(RichMixin):
    """
    Implements the "rich" console rendering of cache information including location,
    size, and dataset statistics.
    """

    def __init__(self,
                 cache_path: str,
                 total_size: int,
                 total_datasets: int,
                 total_files: int,
                 ):
        self.cache_path = cache_path
        self.total_size = total_size
        self.total_datasets = total_datasets
        self.total_files = total_files

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:

        table = Table(title=None, box=None, show_header=False, leading=1, padding=(0, 2), pad_edge=False)
        table.add_column(justify='left', style='magenta', no_wrap=True)
        table.add_column(justify='left', style='white', no_wrap=False)

        # Cache location
        table.add_row('Location', f'[cyan]{self.cache_path}[/cyan]')

        # Total size
        size_mb = self.total_size / (1024 * 1024)
        size_gb = size_mb / 1024
        if size_gb >= 1:
            size_display = f'{size_gb:.2f} GB ({size_mb:.2f} MB)'
        else:
            size_display = f'{size_mb:.2f} MB'
        table.add_row('Total Size', f'[green]{size_display}[/green]')

        # Number of datasets
        table.add_row('Datasets', f'[yellow]{self.total_datasets}[/yellow]')

        # Total files
        table.add_row('Total Files', f'[blue]{self.total_files}[/blue]')

        # Check if cache directory exists
        cache_exists = os.path.exists(self.cache_path)
        status = '[green]✅ Available[/green]' if cache_exists else '[red]❌ Not found[/red]'
        table.add_row('Status', status)

        panel = Panel(
            table,
            title='[bright_black]Cache Information[/bright_black]',
            title_align='left',
            border_style='bright_black',
        )

        yield ''
        yield panel


class RichConfig(RichMixin):
    
    def __init__(self, config_file_path: str):
        self.file_path = config_file_path
        self.file_name = os.path.basename(self.file_path)
        with open(self.file_path) as file:
            self.content = file.read()
        
    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:
            
        rule_top = Rule(
            title=f'[bright_black]{self.file_path}[/bright_black]', 
            align='center', 
            style='bright_black'
        )
        yield rule_top
        
        yield self.content.replace(r'[', r'\[')
            
        rule_bottom = Rule(title=None, style='bright_black', end=' ')
        yield rule_bottom


class RichRemoteInfo(RichMixin):
    """
    Rich display element that shows detailed information about the currently configured 
    remote file share server.
    """
    def __init__(self, config, show_all: bool = False, metadata_available: bool = False, dataset_count: Optional[int] = None):
        self.config = config
        self.show_all = show_all
        self.metadata_available = metadata_available
        self.dataset_count = dataset_count
        
    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:
        
        table = Table(title=None, expand=True, box=None)
        table.add_column("Setting", justify="left", style="magenta", no_wrap=True)
        table.add_column("Value", justify="left", style="white", no_wrap=False)
        
        # Basic configuration
        table.add_row('Fileshare Type', f'[bold cyan]{self.config.get_fileshare_type()}[/bold cyan]')
        table.add_row('Fileshare URL', f'[bold cyan]{self.config.get_fileshare_url()}[/bold cyan]')
        
        # Get fileshare parameters for the specific type
        fileshare_params = self.config.get_fileshare_parameters(self.config.get_fileshare_type())
        
        # DAV Configuration (if available)
        if 'dav_url' in fileshare_params:
            table.add_row('DAV URL', f'[cyan]{fileshare_params["dav_url"]}[/cyan]')
        else:
            table.add_row('DAV URL', '[bright_black]Not configured[/bright_black]')
            
        if 'dav_username' in fileshare_params:
            table.add_row('DAV Username', f'[cyan]{fileshare_params["dav_username"]}[/cyan]')
        else:
            table.add_row('DAV Username', '[bright_black]Not configured[/bright_black]')
        
        # Password handling - show masked unless --all flag is used
        if 'dav_password' in fileshare_params:
            if self.show_all:
                table.add_row('DAV Password', f'[yellow]{fileshare_params["dav_password"]}[/yellow]')
            else:
                table.add_row('DAV Password', '[yellow]***[/yellow] [bright_black](use --all to show)[/bright_black]')
        else:
            table.add_row('DAV Password', '[bright_black]Not configured[/bright_black]')
        
        # Metadata availability
        if self.metadata_available:
            table.add_row('Metadata Available', '[green]✅ Available[/green]')
        else:
            table.add_row('Metadata Available', '[red]❌ Not accessible[/red]')
        
        # Dataset count
        if self.dataset_count is not None:
            table.add_row('Available Datasets', f'[green]{self.dataset_count}[/green]')
        else:
            table.add_row('Available Datasets', '[bright_black]Unknown[/bright_black]')
            
        panel = Panel(
            table,
            title='Remote Configuration',
        )
            
        yield ''
        yield panel
        

class RichDiffDisplay(RichMixin):
    """
    Rich display element that shows the diff between two metadata.yml files.
    """
    def __init__(self, local_file: str, remote_file: str, diff_lines: List[str], changed_lines: int):
        self.local_file = local_file
        self.remote_file = remote_file
        self.diff_lines = diff_lines
        self.changed_lines = changed_lines
        
    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:
        
        yield ''
        
        # Show file paths being compared
        comparison_table = Table(box=None)
        comparison_table.add_column("File", style="magenta", no_wrap=True)
        comparison_table.add_column("Path", style="cyan", no_wrap=False)
        comparison_table.add_row("Local", self.local_file)
        comparison_table.add_row("Remote", "metadata.yml (from server)")
        
        yield comparison_table
        
        # Show the actual diff if there are changes
        if self.changed_lines > 0 and self.diff_lines:
            yield ""
            
            # Limit diff output to first 50 lines to avoid overwhelming display
            display_lines = self.diff_lines[:50]
            if len(self.diff_lines) > 50:
                display_lines.append(f"... ({len(self.diff_lines) - 50} more lines truncated)")
            
            # Create a syntax-highlighted diff
            diff_text = '\n'.join(display_lines)
            syntax = Syntax(diff_text, "diff", theme="monokai", line_numbers=False)
            
            panel = Panel(
                syntax,
                title=f'File Differences (showing first {min(50, len(self.diff_lines))} lines)',
                expand=True
            )
            
            yield panel
            
        # Show diff summary
        yield ""
        if self.changed_lines == 0:
            yield Text(f"✅ Files are identical", style="green")
        else:
            yield Text(f"📊 {self.changed_lines} lines differ between local and remote files", style="yellow")


def calculate_dataset_stats(
    df: pd.DataFrame,
    smiles_column: str,
    sample_size: Optional[int] = None,
    progress: Optional[Progress] = None,
) -> dict:
    """
    Calculate comprehensive statistics for a SMILES dataset.

    This function processes each molecule in the dataset and computes various
    statistics including molecular size, element composition, ring systems,
    drug-likeness metrics, and structural diversity.

    :param df: DataFrame containing SMILES data
    :param smiles_column: Name of the column containing SMILES strings
    :param sample_size: Optional limit on number of molecules to process
    :param progress: Optional Rich Progress instance for progress display

    :returns: Dictionary with statistics organized by category
    """
    # Sample if requested
    if sample_size is not None and sample_size < len(df):
        df = df.sample(n=sample_size, random_state=42)

    total_count = len(df)
    valid_count = 0
    invalid_count = 0

    # Molecule size statistics
    heavy_atom_counts = []
    total_atom_counts = []
    bond_counts = []
    mol_weights = []

    # Element composition
    element_counter: Counter = Counter()

    # Ring statistics
    molecules_with_rings = 0
    aromatic_molecules = 0
    ring_sizes: Counter = Counter()

    # Drug-likeness metrics
    logp_values = []
    hbd_values = []
    hba_values = []
    tpsa_values = []
    rotatable_bond_values = []
    lipinski_passes = 0

    # Structural diversity
    scaffolds: set = set()
    chiral_center_counts = []

    # Functional groups (simplified detection via SMARTS patterns)
    functional_groups: Counter = Counter()
    fg_patterns = {
        'Hydroxyl (-OH)': '[OX2H]',
        'Primary Amine (-NH2)': '[NX3H2]',
        'Carboxylic Acid (-COOH)': '[CX3](=O)[OX2H1]',
        'Aldehyde (-CHO)': '[CX3H1](=O)[#6]',
        'Ketone (C=O)': '[#6][CX3](=O)[#6]',
        'Ester (-COO-)': '[#6][CX3](=O)[OX2H0][#6]',
        'Amide (-CONH-)': '[NX3][CX3](=[OX1])[#6]',
        'Nitro (-NO2)': '[$([NX3](=O)=O),$([NX3+](=O)[O-])]',
        'Sulfonyl (-SO2-)': '[#16X4](=[OX1])(=[OX1])',
        'Halogen (F/Cl/Br/I)': '[F,Cl,Br,I]',
    }
    compiled_patterns = {name: Chem.MolFromSmarts(smarts) for name, smarts in fg_patterns.items()}

    # Progress tracking
    task_id = None
    if progress is not None:
        task_id = progress.add_task("Analyzing molecules...", total=total_count)

    for idx, row in df.iterrows():
        smiles = row[smiles_column]

        try:
            mol = Chem.MolFromSmiles(smiles)
            if mol is None:
                invalid_count += 1
                if progress is not None:
                    progress.update(task_id, advance=1)
                continue

            valid_count += 1

            # Size metrics
            heavy_atoms = Descriptors.HeavyAtomCount(mol)
            heavy_atom_counts.append(heavy_atoms)

            mol_with_h = Chem.AddHs(mol)
            total_atom_counts.append(mol_with_h.GetNumAtoms())
            bond_counts.append(mol.GetNumBonds())
            mol_weights.append(Descriptors.ExactMolWt(mol))

            # Element composition
            for atom in mol.GetAtoms():
                element_counter[atom.GetSymbol()] += 1

            # Ring statistics
            ring_info = mol.GetRingInfo()
            num_rings = ring_info.NumRings()
            if num_rings > 0:
                molecules_with_rings += 1
                for ring in ring_info.AtomRings():
                    ring_sizes[len(ring)] += 1

            if Descriptors.NumAromaticRings(mol) > 0:
                aromatic_molecules += 1

            # Drug-likeness
            logp = Descriptors.MolLogP(mol)
            mw = Descriptors.ExactMolWt(mol)
            hbd = Descriptors.NumHDonors(mol)
            hba = Descriptors.NumHAcceptors(mol)
            tpsa = Descriptors.TPSA(mol)
            rotatable = Descriptors.NumRotatableBonds(mol)

            logp_values.append(logp)
            hbd_values.append(hbd)
            hba_values.append(hba)
            tpsa_values.append(tpsa)
            rotatable_bond_values.append(rotatable)

            # Lipinski Rule of 5
            if mw <= 500 and logp <= 5 and hbd <= 5 and hba <= 10:
                lipinski_passes += 1

            # Scaffold
            try:
                core = MurckoScaffold.GetScaffoldForMol(mol)
                scaffold_smiles = Chem.MolToSmiles(core)
                scaffolds.add(scaffold_smiles)
            except Exception:
                pass

            # Chiral centers
            chiral_centers = Chem.FindMolChiralCenters(mol, includeUnassigned=True)
            chiral_center_counts.append(len(chiral_centers))

            # Functional groups
            for fg_name, pattern in compiled_patterns.items():
                if pattern is not None and mol.HasSubstructMatch(pattern):
                    functional_groups[fg_name] += 1

        except Exception:
            invalid_count += 1

        if progress is not None:
            progress.update(task_id, advance=1)

    # Compute summary statistics
    def safe_stats(values):
        if not values:
            return {'min': 0, 'max': 0, 'mean': 0, 'std': 0}
        arr = np.array(values)
        return {
            'min': float(np.min(arr)),
            'max': float(np.max(arr)),
            'mean': float(np.mean(arr)),
            'std': float(np.std(arr)),
        }

    return {
        'basic': {
            'total_count': total_count,
            'valid_count': valid_count,
            'invalid_count': invalid_count,
            'sampled': sample_size is not None,
        },
        'size': {
            'heavy_atoms': safe_stats(heavy_atom_counts),
            'total_atoms': safe_stats(total_atom_counts),
            'bonds': safe_stats(bond_counts),
            'molecular_weight': safe_stats(mol_weights),
        },
        'elements': dict(element_counter.most_common()),
        'rings': {
            'molecules_with_rings': molecules_with_rings,
            'molecules_with_rings_pct': (molecules_with_rings / valid_count * 100) if valid_count > 0 else 0,
            'aromatic_molecules': aromatic_molecules,
            'aromatic_molecules_pct': (aromatic_molecules / valid_count * 100) if valid_count > 0 else 0,
            'ring_size_distribution': dict(ring_sizes),
        },
        'druglikeness': {
            'logp': safe_stats(logp_values),
            'hbd': safe_stats(hbd_values),
            'hba': safe_stats(hba_values),
            'tpsa': safe_stats(tpsa_values),
            'rotatable_bonds': safe_stats(rotatable_bond_values),
            'lipinski_pass_count': lipinski_passes,
            'lipinski_pass_pct': (lipinski_passes / valid_count * 100) if valid_count > 0 else 0,
        },
        'diversity': {
            'unique_scaffolds': len(scaffolds),
            'scaffold_ratio': (len(scaffolds) / valid_count) if valid_count > 0 else 0,
            'functional_groups': dict(functional_groups),
            'chiral_centers': safe_stats(chiral_center_counts),
        },
    }


class RichDatasetStats(RichMixin):
    """
    Implements the "rich" console rendering of dataset statistics.

    This display element shows comprehensive statistics about a molecular dataset
    organized into multiple panels covering size, composition, rings, drug-likeness,
    and structural diversity.
    """

    def __init__(self, name: str, stats: dict):
        self.name = name
        self.stats = stats

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> t.Any:

        yield ''

        # == Panel 1: Basic Dataset Info ==
        basic = self.stats['basic']
        basic_table = Table(title=None, box=None, show_header=False, padding=(0, 2), expand=True)
        basic_table.add_column(justify='left', style='magenta', no_wrap=True)
        basic_table.add_column(justify='left', style='white', no_wrap=False, ratio=1)

        basic_table.add_row('Dataset', f'[bold cyan]{self.name}[/bold cyan]')
        basic_table.add_row('Total Compounds', str(basic['total_count']))
        basic_table.add_row('Valid SMILES', f"[green]{basic['valid_count']}[/green]")
        if basic['invalid_count'] > 0:
            basic_table.add_row('Invalid SMILES', f"[red]{basic['invalid_count']}[/red]")
        if basic['sampled']:
            basic_table.add_row('', '[bright_black](sampled)[/bright_black]')

        yield Panel(basic_table, title='[bright_black]Basic Info[/bright_black]',
                   title_align='left', border_style='bright_black')

        # == Panel 2: Molecule Size Statistics ==
        size = self.stats['size']
        size_table = Table(box=box.SIMPLE, padding=(0, 1), expand=True)
        size_table.add_column("Metric", style="cyan", no_wrap=True, ratio=2)
        size_table.add_column("Min", justify="right", style="white", ratio=1)
        size_table.add_column("Mean", justify="right", style="green", ratio=1)
        size_table.add_column("Max", justify="right", style="white", ratio=1)
        size_table.add_column("Std", justify="right", style="bright_black", ratio=1)

        size_table.add_row(
            "Heavy Atoms",
            f"{size['heavy_atoms']['min']:.0f}",
            f"{size['heavy_atoms']['mean']:.1f}",
            f"{size['heavy_atoms']['max']:.0f}",
            f"{size['heavy_atoms']['std']:.1f}",
        )
        size_table.add_row(
            "Total Atoms (incl. H)",
            f"{size['total_atoms']['min']:.0f}",
            f"{size['total_atoms']['mean']:.1f}",
            f"{size['total_atoms']['max']:.0f}",
            f"{size['total_atoms']['std']:.1f}",
        )
        size_table.add_row(
            "Bonds",
            f"{size['bonds']['min']:.0f}",
            f"{size['bonds']['mean']:.1f}",
            f"{size['bonds']['max']:.0f}",
            f"{size['bonds']['std']:.1f}",
        )
        size_table.add_row(
            "Molecular Weight",
            f"{size['molecular_weight']['min']:.1f}",
            f"{size['molecular_weight']['mean']:.1f}",
            f"{size['molecular_weight']['max']:.1f}",
            f"{size['molecular_weight']['std']:.1f}",
        )

        yield Panel(size_table, title='[bright_black]Molecule Size[/bright_black]',
                   title_align='left', border_style='bright_black')

        # == Panel 3: Element Composition ==
        elements = self.stats['elements']
        if elements:
            total_atoms = sum(elements.values())
            elem_table = Table(box=box.SIMPLE, padding=(0, 1), expand=True)
            elem_table.add_column("Element", style="cyan", no_wrap=True, ratio=1)
            elem_table.add_column("Count", justify="right", style="white", ratio=1)
            elem_table.add_column("Percentage", justify="right", style="green", ratio=1)

            for elem, count in list(elements.items())[:12]:  # Top 12 elements
                pct = count / total_atoms * 100
                elem_table.add_row(elem, str(count), f"{pct:.1f}%")

            yield Panel(elem_table, title='[bright_black]Element Composition[/bright_black]',
                       title_align='left', border_style='bright_black')

        # == Panel 4: Ring Statistics ==
        rings = self.stats['rings']
        ring_table = Table(box=box.SIMPLE, padding=(0, 1), expand=True)
        ring_table.add_column("Metric", style="cyan", no_wrap=True, ratio=2)
        ring_table.add_column("Count", justify="right", style="white", ratio=1)
        ring_table.add_column("Percentage", justify="right", style="green", ratio=1)

        ring_table.add_row(
            'Molecules with Rings',
            str(rings['molecules_with_rings']),
            f"{rings['molecules_with_rings_pct']:.1f}%"
        )
        ring_table.add_row(
            'Aromatic Molecules',
            str(rings['aromatic_molecules']),
            f"{rings['aromatic_molecules_pct']:.1f}%"
        )

        yield Panel(ring_table, title='[bright_black]Ring Statistics[/bright_black]',
                   title_align='left', border_style='bright_black')

        # Ring size distribution as separate panel
        if rings['ring_size_distribution']:
            total_rings = sum(rings['ring_size_distribution'].values())
            ring_dist_table = Table(box=box.SIMPLE, padding=(0, 1), expand=True)
            ring_dist_table.add_column("Ring Size", style="cyan", no_wrap=True, ratio=1)
            ring_dist_table.add_column("Count", justify="right", style="white", ratio=1)
            ring_dist_table.add_column("Percentage", justify="right", style="green", ratio=1)

            # Sort by count descending and take top 5
            top_ring_sizes = sorted(rings['ring_size_distribution'].items(), key=lambda x: -x[1])[:5]
            for size, count in sorted(top_ring_sizes, key=lambda x: x[0]):  # Display sorted by ring size
                pct = count / total_rings * 100 if total_rings > 0 else 0
                ring_dist_table.add_row(f"{size}-membered", str(count), f"{pct:.1f}%")

            yield Panel(ring_dist_table, title='[bright_black]Ring Size Distribution[/bright_black]',
                       title_align='left', border_style='bright_black')

        # == Panel 5: Drug-likeness Metrics ==
        dl = self.stats['druglikeness']
        dl_table = Table(box=box.SIMPLE, padding=(0, 1), expand=True)
        dl_table.add_column("Property", style="cyan", no_wrap=True, ratio=2)
        dl_table.add_column("Min", justify="right", style="white", ratio=1)
        dl_table.add_column("Mean", justify="right", style="green", ratio=1)
        dl_table.add_column("Max", justify="right", style="white", ratio=1)
        dl_table.add_column("Lipinski", justify="right", style="yellow", ratio=1)

        dl_table.add_row("LogP", f"{dl['logp']['min']:.2f}", f"{dl['logp']['mean']:.2f}",
                        f"{dl['logp']['max']:.2f}", "≤5")
        dl_table.add_row("MW", f"{self.stats['size']['molecular_weight']['min']:.0f}",
                        f"{self.stats['size']['molecular_weight']['mean']:.0f}",
                        f"{self.stats['size']['molecular_weight']['max']:.0f}", "≤500")
        dl_table.add_row("H-Bond Donors", f"{dl['hbd']['min']:.0f}", f"{dl['hbd']['mean']:.1f}",
                        f"{dl['hbd']['max']:.0f}", "≤5")
        dl_table.add_row("H-Bond Acceptors", f"{dl['hba']['min']:.0f}", f"{dl['hba']['mean']:.1f}",
                        f"{dl['hba']['max']:.0f}", "≤10")
        dl_table.add_row("TPSA", f"{dl['tpsa']['min']:.1f}", f"{dl['tpsa']['mean']:.1f}",
                        f"{dl['tpsa']['max']:.1f}", "≤140")
        dl_table.add_row("Rotatable Bonds", f"{dl['rotatable_bonds']['min']:.0f}",
                        f"{dl['rotatable_bonds']['mean']:.1f}",
                        f"{dl['rotatable_bonds']['max']:.0f}", "≤10")

        yield Panel(dl_table, title='[bright_black]Drug-likeness Metrics[/bright_black]',
                   title_align='left', border_style='bright_black')

        # == Panel 6: Structural Diversity ==
        div = self.stats['diversity']
        div_table = Table(title=None, box=None, show_header=False, padding=(0, 2), expand=True)
        div_table.add_column(justify='left', style='magenta', no_wrap=True)
        div_table.add_column(justify='left', style='white', no_wrap=False, ratio=1)

        div_table.add_row('Unique Murcko Scaffolds', str(div['unique_scaffolds']))
        div_table.add_row('Scaffold Diversity Ratio', f"{div['scaffold_ratio']:.3f}")
        div_table.add_row('Avg. Chiral Centers', f"{div['chiral_centers']['mean']:.2f}")

        yield Panel(div_table, title='[bright_black]Structural Diversity[/bright_black]',
                   title_align='left', border_style='bright_black')

        # Functional groups as separate panel
        if div['functional_groups']:
            fg_table = Table(box=box.SIMPLE, padding=(0, 1), expand=True)
            fg_table.add_column("Functional Group", style="cyan", no_wrap=True, ratio=2)
            fg_table.add_column("Molecules", justify="right", style="white", ratio=1)
            fg_table.add_column("Percentage", justify="right", style="green", ratio=1)

            valid_count = self.stats['basic']['valid_count']
            for fg_name, count in sorted(div['functional_groups'].items(), key=lambda x: -x[1]):
                pct = count / valid_count * 100 if valid_count > 0 else 0
                fg_table.add_row(fg_name, str(count), f"{pct:.1f}%")

            yield Panel(fg_table, title='[bright_black]Functional Groups[/bright_black]',
                       title_align='left', border_style='bright_black')


# == ACTUAL CLI IMPLEMENTATION ==
# The following class is the actual implementation of the command line interface for the ChemMatData
# package. This class is a subclass of the RichGroup class from the rich_click package which is a
# subclass of the click.Group class. This class is used to define a group of commands that are all
# related to the same topic. In this case the topic is the ChemMatData package and the commands are
# all related to the downloading and managing of the datasets that are part of the package.

class CLI(click.RichGroup):

    def __init__(self,
                 **kwargs):
        super(CLI, self).__init__(
            invoke_without_command=True,
            **kwargs
        )

        # This config file is a global singleton instance which allows access to the most 
        # important config parameter of the project such as the URL address of the fileshare 
        # server from which the datasets will be downloaded.
        self.config = Config()
        self.file_share = NextcloudFileShare(
            url=self.config.get_fileshare_url(),
            **self.config.get_fileshare_parameters(fileshare_type='nextcloud'),
        )
        self.cache = self.config.cache

        # ~ Constructing the help string.
        # This is the string which will be printed when the help option is called.
        self.rich_help = RichHelp()
        
        # This is the rich object that can be used to print the ASCII art logo to the 
        # console.
        self.rich_logo = RichLogo()
        
        ## -- Adding Actual Commands --
        
        # first level commands
        self.add_command(self.download_command)
        self.add_command(self.list_command)
        self.add_command(self.info_command)
        self.add_command(self.stats_command)

        # cache command group
        self.add_command(self.cache_group)
        self.cache_group.add_command(self.cache_list_command)
        self.cache_group.add_command(self.cache_clear_command)
        self.cache_group.add_command(self.cache_remove_command)
        self.cache_group.add_command(self.cache_info_command)
        
        # config command group
        self.add_command(self.config_group)
        self.config_group.add_command(self.config_show_command)
        self.config_group.add_command(self.config_edit_command)

        # remote command group
        self.add_command(self.remote_group)
        self.remote_group.add_command(self.remote_show_command)
        self.remote_group.add_command(self.remote_diff_command)
        self.remote_group.add_command(self.remote_upload_command)
        self.remote_group.add_command(self.remote_exists_command)

    # Here we override the default "format_help" method of the RichGroup base class.
    # This method is being called to actually render the "--help" option of the command group.
    # We override this here to add the custom behavior of prining the logo before actually
    # printing the help text.
    def format_help(self, ctx: t.Any, formatter: t.Any) -> None:

        # Before printing the help text we want to print the logo
        rich.print(self.rich_logo)

        self.format_usage(ctx, formatter)
        click.echo(self.rich_help)

        self.format_options(ctx, formatter)

        # Custom command display - only call this once
        rich_commands = RichCommands(self.commands)
        click.echo(rich_commands)

        self.format_epilog(ctx, formatter)
        
    def format_commands(self, ctx: Any, formatter: Any) -> None:
        # Override parent behavior - do nothing here since we handle it in format_help
        pass

    # --- commands ---
    # The following methods are actually the command implementations which are the specific commands 
    # that are part of the command group that is represented by the ExperimentCLI object instance itself.

    @click.command('download', short_help='Download a dataset form the remote file share server')
    @click.argument('name')
    @click.option('--full', is_flag=True, help='Download both the original and the processed datasets')
    @click.option('--path', default=os.getcwd(), type=click.Path(file_okay=False), help='Path to where the files will be downloaded')
    @click.pass_obj
    # Try with "clintox"
    def download_command(self, name: str, full: bool, path: str):
        """
        Downloads the dataset with the given NAME from the remote file share server to the local system. 
        """
        click.secho('Connecting to server...')
        self.file_share.fetch_metadata(force=True)
        
        # First of all we can check if the dataset is even registered in the metadata file 
        # from the file share server. If that is already not the case we dont even need to 
        # try and fetch the dataset files itself.
        if name not in self.file_share['datasets']:
            click.secho('Dataset not found! Use the "list" command to see available datasets...', fg='red')
            return 1
        
        dataset_metadata: dict = self.file_share['datasets'][name]
        
        with Progress() as progress:
            
            click.secho('Downloading raw dataset...')
            for file_extension in dataset_metadata['raw']:
                file_name = f'{name}.{file_extension}'

                # First try to download the gzipped version
                try:
                    file_name_compressed = f'{file_name}.gz'
                    file_path_compressed = self.file_share.download_file(
                        file_name_compressed,
                        folder_path=path,
                        progress=progress,
                    )

                    # Decompress the file
                    file_path = os.path.join(path, file_name)
                    with open(file_path, mode='wb') as file:
                        with gzip.open(file_path_compressed, mode='rb') as compressed_file:
                            shutil.copyfileobj(compressed_file, file)

                    # Remove the compressed file after decompression
                    os.remove(file_path_compressed)

                # If gzipped version fails, download uncompressed version
                except Exception:
                    file_path = self.file_share.download_file(
                        file_name,
                        folder_path=path,
                        progress=progress
                    )

                self.cache.add_dataset(
                    name=name,
                    typ=file_extension,
                    path=file_path,
                    metadata=dataset_metadata,
                )
                
            # We only want to download the full dataset (aka the message packed version of the 
            # graph dicts) when the corresponding flag is explicitly set for the command
            if full:
                
                # There is also the possibility that a full dataset is not even available for 
                # the dataset in question. In that case we want to inform the user about that
                if not dataset_metadata['full']:
                    click.secho('Full format dataset not available!', fg='red')
                
                path = self.file_share.download_dataset(
                    f'{name}.mpack', 
                    folder_path=path, 
                    progress=progress
                )
                
                # 01.11.24
                # After downloading the full dataset we want to store that dataset in the cache for 
                # future use. This way we can avoid downloading the dataset again in the future.
                # After each download we want to replace the version in the cache with the new one.
                self.cache.add_dataset(
                    name=name,
                    typ='mpack',
                    path=path,
                    metadata=dataset_metadata,
                )
        
        click.secho('Download complete!', fg='green')
        
    @click.command('list', short_help='List the available datasets')
    @click.option('-s', '--sort', is_flag=True, help='Sort the datasets by name.')
    @click.option('--show-hidden', is_flag=True, help='Show hidden datasets as well. Mainly for testing purposes')
    @click.pass_obj
    def list_command(self, sort: bool, show_hidden: bool):
        """
        This command will display a list of the available datasets that are available for download.
        The list will also display some basic information about each dataset such as the number of
        elements, the number of targets, the target type and some tags.
        """
        
        click.secho('Connecting to server...')
        self.file_share.fetch_metadata(force=True)
        
        rich_list = RichDatasetList(
            self.file_share['datasets'],
            sort=sort,
            show_hidden=show_hidden,    
        )
        click.echo(rich_list)
        
        return 0
           
    @click.command('info', short_help='Show detailed information about one of the datasets')
    @click.argument('name')
    @click.pass_obj
    def info_command(self, name: str):
        """
        This command will print detailed information about the dataset with the given NAME.
        """
        self.file_share.fetch_metadata()
        if name not in self.file_share['datasets']:
            click.secho('Dataset not found on the remote file share server!', fg='red')
            return
        
        dataset_info: dict = self.file_share['datasets'][name]
        rich_info = RichDatasetInfo(name, dataset_info)
        click.echo(rich_info)

        return 0

    @click.command('stats', short_help='Show statistical analysis of a dataset')
    @click.argument('name')
    @click.option('--sample', '-s', type=int, default=None,
                  help='Sample N molecules for analysis (default: all)')
    @click.pass_obj
    def stats_command(self, name: str, sample: Optional[int]):
        """
        Download and analyze a dataset, showing molecular statistics.

        This command downloads the specified dataset to a temporary folder,
        calculates comprehensive statistics about the molecules, and displays
        them in a formatted output. Statistics include molecule sizes, element
        composition, ring systems, drug-likeness metrics (Lipinski Rule of 5),
        and structural diversity measures.

        Use the --sample flag to limit analysis to a random subset of molecules
        for faster processing on large datasets.
        """
        click.secho('Connecting to server...')
        self.file_share.fetch_metadata(force=True)

        # Check if the dataset exists
        if name not in self.file_share['datasets']:
            click.secho('Dataset not found! Use the "list" command to see available datasets...', fg='red')
            return 1

        dataset_metadata: dict = self.file_share['datasets'][name]

        # Download to temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            click.secho(f'Downloading dataset to temporary folder...')

            # Find the SMILES column name - datasets use different conventions
            smiles_column = None

            with Progress() as progress:
                # Download the raw CSV dataset
                for file_extension in dataset_metadata.get('raw', []):
                    if file_extension == 'csv':
                        file_name = f'{name}.{file_extension}'

                        # Try gzipped version first
                        try:
                            file_name_compressed = f'{file_name}.gz'
                            file_path_compressed = self.file_share.download_file(
                                file_name_compressed,
                                folder_path=temp_dir,
                                progress=progress,
                            )

                            # Decompress the file
                            file_path = os.path.join(temp_dir, file_name)
                            with open(file_path, mode='wb') as file:
                                with gzip.open(file_path_compressed, mode='rb') as compressed_file:
                                    shutil.copyfileobj(compressed_file, file)

                            os.remove(file_path_compressed)

                        except Exception:
                            file_path = self.file_share.download_file(
                                file_name,
                                folder_path=temp_dir,
                                progress=progress
                            )

                        break
                else:
                    click.secho('No CSV format available for this dataset!', fg='red')
                    return 1

            # Load the dataset
            click.secho('Loading dataset...')
            df = pd.read_csv(file_path)

            # Try to find the SMILES column
            possible_smiles_columns = ['smiles', 'SMILES', 'Smiles', 'Canonical_Smiles',
                                       'canonical_smiles', 'mol', 'molecule']
            for col in possible_smiles_columns:
                if col in df.columns:
                    smiles_column = col
                    break

            if smiles_column is None:
                # Try to find any column containing 'smiles' in its name
                for col in df.columns:
                    if 'smiles' in col.lower():
                        smiles_column = col
                        break

            if smiles_column is None:
                click.secho(f'Could not find SMILES column. Available columns: {list(df.columns)}', fg='red')
                return 1

            # Calculate statistics
            click.secho(f'Analyzing {len(df)} molecules...')
            with Progress() as progress:
                stats = calculate_dataset_stats(
                    df=df,
                    smiles_column=smiles_column,
                    sample_size=sample,
                    progress=progress,
                )

            # Display results
            rich_stats = RichDatasetStats(name, stats)
            click.echo(rich_stats)

        return 0

    @click.command('about', short_help='print additional information about the command line interface')
    @click.pass_obj
    def about(self):
        # TODO: Implement an "about" page which prints additional information about the command line interface
        pass
    
    # ~ CACHE COMMAND GROUP
    # The following methods are the implementations of the commands that are part of the cache command group 
    # which can be used to manage the local file system cache.
    
    @click.group('cache', help='Commands for managing the local dataset cache')
    @click.pass_obj
    def cache_group(self):
        """
        Commands for managing the local dataset cache. When downloading datasets from the remote cache they 
        will be stored in a local cache directory to avoid downloading the same dataset multiple times. This 
        command group allows to view & manipulate the data stored in this cache.
        """
        pass
    
    @click.command('list', short_help='List the datasets that are currently stored in the cache')
    @click.pass_obj
    def cache_list_command(self):
        """
        Shows a list of all the datasets that are currently stored in the cache along with some of the metadata
        that is stored along with the dataset.
        """
        if len(self.cache) == 0:
            click.secho('Cache is empty!', fg='yellow')

        else:
            click.echo('')

            # This method will return all the datasets that are stored in the cache as a list of tuples where the
            # first element of the tuple is the name of the dataset and the second element is the type of the dataset.
            datasets: List[Tuple[str, str]] = self.cache.list_datasets()
            dataset_metadata_map: Dict[Tuple[str, str], dict] = {}
            dataset_sizes: Dict[Tuple[str, str], int] = {}

            for name, typ in datasets:
                # Given the name and type of dataset, this function will simply read the metadata from the
                # corresponding yml file.
                metadata = self.cache.get_dataset_metadata(name, typ)
                dataset_metadata_map[(name, typ)] = metadata

                # Get the size of each dataset
                size_bytes = self.cache.get_dataset_size(name, typ)
                dataset_sizes[(name, typ)] = size_bytes

            # Get total cache size
            total_cache_size = self.cache.get_total_cache_size()

            rich_cache_list = RichCacheList(
                dataset_metadata_map,
                dataset_sizes,
                total_cache_size,
                sort=True
            )
            click.echo(rich_cache_list)
            
    @click.command('clear', short_help='Clear the entire cache')
    @click.option('-v', '--verbose', is_flag=True, help='print verbose output')
    @click.pass_obj
    def cache_clear_command(self, verbose: bool):
        """
        Clears the local cache directory which contains all the datasets that have been downloaded from the
        remote file share server.
        """
        click.echo('clearing cache...')
        num_elements: int = 0
        for file_name, file_path in self.cache.iterator_clear_():
            num_elements += 1
            click.secho(f' > remove "{file_name}"...', fg='bright_black')
            
        click.secho(f'cleared {num_elements} elements', fg='green')

    @click.command('remove', short_help='Remove a specific dataset from the cache')
    @click.argument('name')
    @click.option('--type', '-t', default='mpack', help='Type of dataset to remove (default: mpack)')
    @click.pass_obj
    def cache_remove_command(self, name: str, type: str):
        """
        Removes a specific dataset from the cache by NAME and optionally TYPE.

        The NAME should be the dataset identifier (e.g., 'clintox').
        The TYPE specifies the format of the dataset (e.g., 'mpack', 'csv').
        """
        if self.cache.contains_dataset(name, type):
            success = self.cache.remove_dataset(name, type)
            if success:
                click.secho(f'Successfully removed "{name}.{type}" from cache', fg='green')
            else:
                click.secho(f'Failed to remove "{name}.{type}" from cache', fg='red')
                return 1
        else:
            click.secho(f'Dataset "{name}.{type}" not found in cache', fg='yellow')

            # Show available datasets for helpful suggestions
            datasets = self.cache.list_datasets()
            if datasets:
                click.echo('\nAvailable datasets in cache:')
                for dataset_name, dataset_type in datasets:
                    click.echo(f'  - {dataset_name}.{dataset_type}')
            else:
                click.echo('Cache is empty.')
            return 1

    @click.command('info', short_help='Show information about the cache')
    @click.pass_obj
    def cache_info_command(self):
        """
        Displays detailed information about the cache including location, size,
        and statistics about the datasets stored in it.
        """
        # Get cache statistics
        total_size = self.cache.get_total_cache_size()
        datasets = self.cache.list_datasets()
        total_datasets = len(datasets)

        # Count total files in cache directory
        total_files = 0
        if os.path.exists(self.cache.path):
            for file_name in os.listdir(self.cache.path):
                file_path = os.path.join(self.cache.path, file_name)
                if os.path.isfile(file_path):
                    total_files += 1
                elif os.path.isdir(file_path):
                    # Count files in subdirectories too
                    for root, dirs, files in os.walk(file_path):
                        total_files += len(files)

        rich_cache_info = RichCacheInfo(
            cache_path=self.cache.path,
            total_size=total_size,
            total_datasets=total_datasets,
            total_files=total_files
        )
        click.echo(rich_cache_info)

    ## == CONFIG COMMAND GROUP ==
    # The following methods are the implementations of the commands that are part of the config command group
    # which can be used to manage the config file of the ChemMatData installation.
    
    @click.group('config', help='Commands for managing the configuration file')
    @click.pass_obj
    def config_group(self):
        pass
    
    @click.command('show', short_help='Show the current configuration settings')
    @click.pass_obj
    def config_show_command(self):
        """
        Prints the contents of the config file to the console.
        """
        rich_config = RichConfig(self.config.config_file_path)
        click.echo(rich_config, nl=False)
        
    @click.command('edit', short_help='Edit the configuration file')
    @click.pass_obj
    def config_edit_command(self):
        open_file_in_editor(self.config.config_file_path)

    ## == REMOTE GROUP ==
    # This command group exposes specific commands to interact with the remote file share server
    # which are more targeted towards the maintainers of the database to expose some utility 
    # functions that are not really meant to be used by the end users of the package.
    
    @click.group('remote', help='Commands for interacting with the remote file share server')
    @click.pass_obj
    def remote_group(self):
        pass
    
    def check_dav_credentials(self) -> None:
        """
        Checks if the DAV credentials are set up correctly in the config file and if not prints a 
        warning message to the user.
        
        :raises: AssertionError if the DAV credentials are not set up correctly
        
        :returns: None
        """
        try:
            self.file_share.assert_dav()
        except AssertionError:
            click.secho('⚠️ DAV credentials are not set up!'
                        'You need to add valid DAV credentials in the config file to interact '
                        'with files located on the remote file share server.')
    
    @click.command('upload', short_help='Upload a dataset or file to the remote file share server')
    @click.argument('file_path', type=click.Path(exists=True, dir_okay=False))
    @click.option('--name', default=None, help=(
        'The name of the file on the remote server. If not provided, the name of the file '
        'itself will be used.'
    ))
    @click.pass_obj
    def remote_upload_command(self, 
                              file_path: str,
                              name: Optional[str],
                              **kwargs
                              ) -> None:
        
        ## -- Checking Credentials --
        # This will make sure that the DAV credentials are set up correctly before trying to upload 
        # anything. If the credentials are missing, this will print to the user.
        self.check_dav_credentials()

        ## -- Uploading the file --
        if name is not None:
            file_name: str = name
        else: # If the name is not provided, we will use the name of the file itself
            file_name = os.path.basename(file_path)
        
        file_size: int = os.path.getsize(file_path)
        file_size_mb: int = file_size // (1024 * 1024)
        click.secho(f'... Uploading file "{file_name}" with {file_size_mb} MB')
        
        # This method will actually upload the file to the remote server.
        self.file_share.upload(file_name, file_path)

        ## -- Checking for the file --
        # After the upload is complete, we want to check if the file was actually uploaded.
        exists, meta = self.file_share.exists(file_name)
        if exists:
            click.secho('✅ File uploaded successfully!')
        else:
            click.secho('⚠️ File upload failed!')

    @click.command('exists', short_help='Check if a file exists on the rmote file share server')
    @click.argument('file_name')
    @click.pass_obj
    def remote_exists_command(self,
                              file_name: str,
                              ) -> None:
        """
        Checks if the file with the given FILE_NAME exists on the remote file share server.
        """
        ## -- Checking Credentials --
        # This will make sure that the DAV credentials are set up correctly before trying to upload
        # anything. If the credentials are missing, this will print to the user.
        self.check_dav_credentials()
        
        ## -- Checking for the file --
        # This method will check if the file with the given name exists on the remote server.
        exists, meta = self.file_share.exists(file_name)
        if exists:
            click.secho(f'✅ File "{file_name}" exists on the remote server!')
            click.secho(f'  - Size: {meta["size"]} bytes')
            click.secho(f'  - Last modified: {meta["last_modified"]}')
            click.secho(f'  - Content type: {meta["content_type"]}')
            
        else:
            click.secho(f'⚠️ File "{file_name}" does not exist on the remote server!', fg='red')
            sys.exit(1)
    
    @click.command('show', short_help='Display remote configuration information')
    @click.option('--all', is_flag=True, help='Show all information including sensitive data like passwords')
    @click.pass_obj
    def remote_show_command(self, all: bool) -> None:
        """
        Displays detailed information about the currently configured remote file share server,
        including URLs, DAV configuration, and metadata availability.
        """
        
        ## -- Check Metadata Availability --
        metadata_available = False
        dataset_count = None
        
        try:
            # Try to fetch metadata to see if it exists
            self.file_share.fetch_metadata(force=True)
            metadata_available = True
            
            # Get dataset count if available
            if 'datasets' in self.file_share.metadata:
                dataset_count = len(self.file_share.metadata['datasets'])
                
        except Exception:
            # Metadata is not accessible
            metadata_available = False
            dataset_count = None
        
        ## -- Creating Rich Display --
        rich_remote_info = RichRemoteInfo(
            self.config, 
            show_all=all, 
            metadata_available=metadata_available, 
            dataset_count=dataset_count
        )
        click.echo(rich_remote_info)
    
    @click.command('diff', short_help='Compare local metadata.yml with remote version')
    @click.argument('local_file', type=click.Path(exists=True, dir_okay=False, readable=True))
    @click.pass_obj
    def remote_diff_command(self, local_file: str) -> None:
        """
        Compares a local metadata.yml file with the remote metadata.yml file from the server.
        Shows the number of different lines and displays a visual diff of the changes.
        
        LOCAL_FILE: Path to the local metadata.yml file to compare
        """
        
        ## --- Fetch Remote Metadata ---
        try:
            # Try to fetch the remote metadata
            self.file_share.fetch_metadata(force=True)
            
            if not hasattr(self.file_share, 'metadata') or self.file_share.metadata is None:
                click.secho('❌ Could not fetch remote metadata.yml file', fg='red')
                sys.exit(1)
                
        except Exception as e:
            click.secho(f'❌ Error fetching remote metadata: {str(e)}', fg='red')
            sys.exit(1)
        
        ## --- Read Local File ---
        try:
            with open(local_file, 'r', encoding='utf-8') as f:
                local_content = f.read()
        except Exception as e:
            click.secho(f'❌ Error reading local file: {str(e)}', fg='red')
            sys.exit(1)
        
        ## --- Normalize Both Files for Comparison ---
        import yaml
        try:
            # Parse local file to normalize it
            local_data = yaml.safe_load(local_content)
            # Convert both to the same YAML format for fair comparison
            local_normalized = yaml.dump(local_data, default_flow_style=False, sort_keys=True, indent=2)
            remote_normalized = yaml.dump(self.file_share.metadata, default_flow_style=False, sort_keys=True, indent=2)
        except Exception as e:
            click.secho(f'❌ Error normalizing YAML data: {str(e)}', fg='red')
            sys.exit(1)
        
        ## --- Calculate Diff ---
        local_lines = local_normalized.splitlines(keepends=False)
        remote_lines = remote_normalized.splitlines(keepends=False)
        
        # Generate unified diff
        diff_lines = list(difflib.unified_diff(
            local_lines,
            remote_lines,
            fromfile=f'local/{os.path.basename(local_file)}',
            tofile='remote/metadata.yml',
            lineterm=''
        ))
        
        # Count changed lines (lines starting with + or - but not +++ or ---)
        changed_lines = sum(1 for line in diff_lines 
                           if (line.startswith('+') and not line.startswith('+++')) or
                              (line.startswith('-') and not line.startswith('---')))
        
        ## --- Display Results ---
        rich_diff = RichDiffDisplay(
            local_file=local_file,
            remote_file="remote/metadata.yml",
            diff_lines=diff_lines,
            changed_lines=changed_lines
        )
        click.echo(rich_diff)


@click.group(cls=CLI)
@click.option("-v", "--version", is_flag=True, help='Print the version of the package')
@click.pass_context
def cli(ctx: t.Any,
        version: bool
        ) -> None:

    # 07.06.24
    # This is actually required to make the CLI class work as it is currently implemented.
    # The problem is that all the commands which are also methods need to be decorated with the 
    # @pass_obj in the position of the "self" argument and here we set the object to be passed 
    # as the CLI instance to make that happen. 
    ctx.obj = ctx.command

    if version:
        version = get_version()
        click.echo(Text(version, style=Style(bold=True)))
        sys.exit(0)
        
    # 07.06.24
    # There was previously a bug here which caused the help text to be rendered for every 
    # command. After adding the additional condition "not ctx.invoked_subcommand" this was 
    # fixed.
    
    # This section will implement the small convenience feature that if no arguments are passed 
    # to the CLI at all, it will simply show the help information.
    elif not ctx.args and not ctx.invoked_subcommand:
        ctx.command.format_help(ctx, ctx.formatter_class())



if __name__ == "__main__":
    cli()  # pragma: no cover



"""
from typing import List, Dict
This module implements the conversion of a dataset of molecules (SMILES strings) to 
fully pre-processed dataset of GraphDict object in message pack format that can be 
directly loaded for the training of graph neural networks.

This module heavily makes use of parallel processing to speed up the conversion 
process. In the main process, the raw data of the molecule elements is fed into a 
multiprocessing queue which is then consumed by several worker processes that 
use the ``MoleculeProcessing`` instance to turn the SMILES string of the molecule 
into a full GraphDict representation. These graph dicts are sent back to the 
main process where they are collected and then later saved into a message pack 
file in the experiments results folder. As an additional optimization those 
message pack files are then compressed into a gzip file to reduce the file size
even further.

**Base Experiment**

This module is actually just the base experiment implementation that should NOT be 
modified directly. The implementation of processing *specific* datasets should only 
be done by creating new sub experiments which inherit the main functionality from 
this base experiment. This can be done by using the ``Experiment.extend`` method.

These sub experiment can inject custom code / functionality to the processing process 
by using the various hooks that are defined in the base experiment.
"""
import os
import time
import csv
import gzip
import shutil
import datetime
import multiprocessing
from typing import Union, List, Dict

import msgpack
import numpy as np
import rdkit.Chem as Chem
import pandas as pd
from rich.pretty import pprint
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

import chem_mat_data._typing as typc
from chem_mat_data.processing import MoleculeProcessing
from chem_mat_data.data import default, ext_hook
from chem_mat_data.data import save_graphs
from chem_mat_data.utils import SCRIPTS_PATH
import yaml

# == SOURCE PARAMETERS ==
# These global parameters can be used to configure the source files from which the 
# dataset will be created. In the default implementation, this source file is assumed 
# to be a CSV file which contains at least the molecule smiles representation and 
# the target values.

# :param SOURCE_PATH:
#       This is the absolute string path of the dataset CSV file from which the graph
#       dataset should be created. 
SOURCE_PATH: str = os.path.join(SCRIPTS_PATH, 'assets', '_test.csv')
# :param SMILES_COLUMN:
#       This is the string name of the CSV column which contains the SMILES strings of
#       the molecules.
SMILES_COLUMN: str = 'smiles'
# :param TARGET_COLUMNS:
#       This is a list of string names of the CSV columns which contain the target values
#       of the dataset. This can be a single column for regression tasks or multiple columns
#       for multi-target regression or classification tasks. For the final graph dataset
#       the target values will be merged into a single numeric vector that contains the 
#       corresponding values in the same order as the column names are defined here.
TARGET_COLUMNS: List[str] = ['target']
# :param DATASET_TYPE:
#       Either 'regression' or 'classification' to define the type of the dataset. This
#       will also determine how the target values are processed.
DATASET_TYPE: str = 'regression'
# :param DESCRIPTION:
#       This is a string description of the dataset that will be stored in the experiment
#       metadata.
DESCRIPTION: str = 'the description of the dataset'
# :param METADATA:
#       A dictionary which will be used as the basis for the metadata that will be added 
#       as additional information to the file share server.
METADATA: dict = {
    'tags': ['Molecules']
}

# == PROCESSING PARAMETERS ==
# These parameters can be used to configure the processing functionality of the script 
# itself. This includes for example whether or not the molecule coordinates should be 
# created by RDKIT as well or not. Also the compression of the final dataset file can
# be configured here.

# :param DATASET_NAME:
#       This string determines the name of the message pack dataset file that is then 
#       stored into the "results" folder of the experiment as the result of the 
#       processing process. The corresponding file extensions will be added 
#       automatically.
DATASET_NAME: str = 'dataset'
# :param USE_COORDINATES:
#       If this is True, then the graph representations of the molecules will also contain 
#       the "node_coordinates" field which will be populated with the 3D coordinates of
#       of the nodes created by RDKIT. Setting this to True will increase the processing 
#       runtime and may cause the processing of some molecules to fail entirely.
USE_COORDINATES: bool = False
# :param COMPRESS:
#       If this is True, then the final message pack dataset file will be compressed into
#       a gzip file to reduce the file size. This is recommended because the raw message
#       pack files are still relatively large. However, the compression will likely take 
#       a while to complete. 
COMPRESS: bool = True

# == EXPERIMENT PARAMETERS ==

# :param __DEBUG__:
#       In debug mode, the experiment will not create a unique results folder but instead will
#       overwrite the results of the last experiment. This is useful for debugging the experiment
#       code itself.
__DEBUG__ = True
# :param __TESTING__:
#       If this is True, then the experiment will be started in "testing" mode 
__TESTING__ = False

experiment = Experiment(
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)


class ProcessingWorker(multiprocessing.Process):
    """
    This class implements the actual processing of the SMILES representation to the 
    graph representation. It inheritted from the multiprocessing.Process class which means 
    that it will be run in a different process which enables multiple molecules to be 
    processed in parallel.
    
    Communication bween the main process and this worker process is done via two
    multiprocessing queues. The input queue is used to send the raw data to be processed 
    to the worker process and the output queue is used to send the processed graph 
    representations back to the main process.
    """
    
    def __init__(self,
                 input_queue: multiprocessing.Queue,
                 output_queue: multiprocessing.Queue,
                 *args,
                 ):
        super(ProcessingWorker, self).__init__()

        self.input_queue = input_queue
        self.output_queue = output_queue

        # This is the object that handles the actual processing of the SMILES.
        self.processing = MoleculeProcessing()
        
    def run(self):
        """
        This code will be executed in a separate process. It essentially consists of an 
        infinite loop that waits for new data from the input queue, then attempts to process 
        the smiles contained in that input data into a graph and then sends the graph back 
        to the main process via the output queue.
        
        It is possible that the processing of molecule fails due to some reason. In this case, 
        the process will put a "None" value into the output queue instead of the graph object.
        The important thing is that for every input element there will be exactly one output.
        """
        for data in iter(self.input_queue.get, None):
            
            try:
                # 10.12.24 - If we want to support the processing of xyz file based datasets as well we need 
                # to not only support the conversion based on a smiles string representation but also the 
                # direct conversion based from a Mol object directly. The processing.process method supports 
                # this inherently to use either - we only have to check here what we are actually getting.
                value: Union[Chem.Mol, str]
                if 'mol' in data:
                    value = data['mol']
                else:
                    value = data['smiles']
                
                graph: typc.GraphDict = self.processing.process(
                    value=value, 
                    use_node_coordinates=experiment.USE_COORDINATES,
                )
                
                graph['graph_labels'] = np.array(data['targets'])
                experiment.apply_hook(
                    'add_graph_metadata',
                    data=data,
                    graph=graph,
                )
                
            except Exception as exc:
                print(f' ! error processing {value} - {exc.__class__.__name__}: {exc}')
                graph = None
            
            graph_encoded = msgpack.packb(graph, default=default)            
            self.output_queue.put(graph_encoded)


@experiment.hook('add_graph_metadata', default=True, replace=False)
def add_graph_metadata(e: Experiment,
                       data: dict,
                       graph: typc.GraphDict
                       ) -> dict:
    """
    This hook is invoked in the processing worker after the SMILES code has been converted 
    to the graph dict already. The hook receives the original data dict and the graph dict 
    as arguments and provides the opportunity to add additional metadata to the graph dict.
    
    ---
    
    This hook is not used by default.
    """
    pass


@experiment.hook('load_dataset', default=True, replace=False)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    """
    In the experiment, this hook is invoked at the very beginning to obtain the actual 
    raw data of the dataset that should be processed. The output of this function should 
    be a dictionary whose keys are the integer indices of the data elements and the values 
    are in turn dictionary objects that should contain AT LEAST the following keys:
    - 'smiles': The SMILES representation of the molecule
    - 'targets': A list of float target values for the molecule

    ---

    This default implementation assums that the dataset is stored in a CSV file and that
    the SMILES representation of the molecules is stored in a column name determined by 
    SMILES_COLUMN and that the target values are stored in columns determined by TARGET_COLUMNS.    

    This default implementation may be overwritten in sub experiments to achieve a more 
    customized behavior.
    """
    if os.path.exists(e.SOURCE_PATH) and os.path.isfile(e.SOURCE_PATH):
        
        # ~ reading csv file
        dataset = {}
        with open(e.SOURCE_PATH) as file:
            dict_reader = csv.DictReader(file)
            for index, row in enumerate(dict_reader):
                dataset[index] = row
                
        # ~ constructing target values
        for index, data in dataset.items():
            targets = []
            
            if e.DATASET_TYPE == 'regression':
                for key in e.TARGET_COLUMNS:
                    targets.append(data[key])
            
            data['targets'] = targets
            data['smiles'] = data[e.SMILES_COLUMN]
            
        return dataset
    
    e.log(f'dataset @ "{e.SOURCE_PATH}" not found!')


@experiment.hook('save_csv', default=True, replace=False)
def save_csv(
    e: Experiment, 
    dataset: Dict[int, dict]
) -> None:
    """
    This hook is invoked right after the dataset has been loaded and processed. It is supposed 
    to save the raw dataset as a CSV file into the experiment archive folder.
    """
    
    # constructing a dataframe from the dataset dict structure
    e.log('constructing dataframe from dataset dict...')
    # Construct a list of dictionaries for DataFrame creation
    df_data = []
    for index, data in dataset.items():
        row = {
            'smiles': data['smiles']
        }
        # Add each target as a separate column
        for i, target in enumerate(data['targets']):
            if len(e.TARGET_COLUMNS) > i:
                row[e.TARGET_COLUMNS[i]] = target
            else:
                row[f'target_{i}'] = target
        df_data.append(row)
    
    df = pd.DataFrame(df_data)
        
    ## -- Save Dataset --
    e.log('Saving the dataset as CSV and GZipped CSV file...')
    csv_path = os.path.join(e.path, f'{e.DATASET_NAME}.csv')
    df.to_csv(csv_path, index=False)

    gz_path = csv_path + '.gz'
    with open(csv_path, 'rb') as f_in, gzip.open(gz_path, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)


@experiment
def experiment(e: Experiment):
    """
    This is the main experiment code that is executed when the script is run.
    """
    
    e.log('starting experiment...')
    e.log_parameters()
    
    # We need to setup the multiprocessing at the beginning here because of the following 
    # technical problem:
    # When python spawns a subprocess it doesnt actually spawn a completely blank python runtime
    # but it *copies* the current process. This also means that it copies the complete memory 
    # of the current process. If we started these child processes after loading the dataset we'd 
    # be copying the entire dataset X times which could very easily lead to a memory overflow...
    e.log('creating processing workers...')
    input_queue = multiprocessing.Queue(maxsize=100)
    output_queue = multiprocessing.Queue(maxsize=100)
    workers = []
    for _ in range(os.cpu_count()):
        worker = ProcessingWorker(
            input_queue=input_queue,
            output_queue=output_queue,
        )
        worker.start()
        workers.append(worker)
        e.log(' * started worker')

    try:
        e.log('loading dataset...')
        dataset: Dict[int, dict] = e.apply_hook(
            'load_dataset'
        )
        num_elements = len(dataset)
        e.log(f'loaded dataset with {num_elements} elements')

        # This following hook will save the dataset as a csv file to the experiment folder.
        e.apply_hook(
            'save_csv',
            dataset=dataset,
        )

        # when starting the experiment in testing mode we want to limit the number of elements in
        # the dataset to only a few examples so that the experiment overall finishes really quickly
        # and so that all the code in the experiment (from start to finish) can be tested quickly
        # for obvious errors (e.g. syntax errors, etc.)
        if e.__TESTING__:

            e.log('running experiment in testing mode...')

            e.log(' * limiting to only 50 elements')
            num_elements = min(num_elements, 50)
            dataset = dict(list(dataset.items())[:num_elements])

        e.log('processing dataset...')
        indices = list(dataset.keys())
        num_indices = len(indices)
        graphs = []

        start_time = time.time()
        count = 0
        prev_count = 0

        # The actual processing happens in the additional worker processes and in this loop we simply
        # feed the data to be processed into the input queue for the workers processes and then collect
        # the processed graphs from the output queue and store that in the local "graphs" list.
        # In the end all the graph dicts in that graphs list will be saved into a msgpack file to the
        # disk.
        while count < num_indices:

            # In each iteration we will completely fill up the input queue so that it is always full
            # and the workers are always busy
            while not input_queue.full() and len(indices) != 0:
                index = indices.pop()
                data = dataset[index]
                input_queue.put(data)

            # Then we also want to clear all the contents of output queue and transfer them to the
            # local graph list
            while not output_queue.empty():
                graph_encoded = output_queue.get()
                graph = msgpack.unpackb(graph_encoded, ext_hook=ext_hook)
                if graph:
                    graphs.append(graph)

                count += 1

            if count % 1000 == 0 and count != prev_count:
                prev_count = count
                time_passed = time.time() - start_time
                num_remaining = num_elements - (count + 1)
                time_per_element = time_passed / (count + 1)
                time_remaining =time_per_element * num_remaining
                eta = datetime.datetime.now() + datetime.timedelta(seconds=time_remaining)
                e.log(f' * {count:05d}/{num_elements} done'
                      f' - time passed: {time_passed / 60:.2f}m'
                      f' - time remaining: {time_remaining / 60:.2f}m'
                      f' - eta: {eta:%a %d.%m %H:%M}')

        end_time = time.time()
        duration = end_time - start_time
        e.log(f'finished processing dataset after {duration / 3600:.2f}h')

    finally:
        # We need to stop the subprocesses here because if we don't do that then the main process won't be
        # able to properly exit either. This is done in a finally block to ensure workers are cleaned up
        # even if an exception occurs during processing.
        e.log('stopping the workers...')
        for worker in workers:
            input_queue.put(None)
            worker.terminate()
            worker.join()

        del input_queue
        del output_queue

    e.log('Description:')
    e.log(e.DESCRIPTION)

    # ~ saving the dataset into the msgpack format
    e.log(f'saving the dataset with {len(graphs)} graphs...')
    dataset_path = os.path.join(e.path, e.DATASET_NAME + '.mpack')
    save_graphs(graphs, dataset_path)
    
    file_size = os.path.getsize(dataset_path)
    file_size_mb = file_size / (1024 * 1024)
    e.log(f'wrote file with {file_size_mb:.1f} MB')
    
    # ~ compressing the file
    # The raw message pack files are actually still relatively large (surprisingly larger than JSON)
    # so we additionally compress those files into a ZIP here so that we can then put those zipped 
    # files onto the server to be actually downloaded. This should greatly reduce the download size 
    # and therefore speed by another 10x or so.
    if e.COMPRESS:
        
        e.log('compressing the dataset file, this may take a while...')
        compressed_path = os.path.join(e.path, e.DATASET_NAME + '.mpack.gz')
        with open(dataset_path, mode='rb') as file:
            with gzip.open(compressed_path, mode='wb') as compressed_file:
                shutil.copyfileobj(file, compressed_file)
                
        compressed_size = os.path.getsize(compressed_path)
        compressed_size_mb = compressed_size / (1024 * 1024)
        e.log(f'compressed file with {compressed_size_mb:.1f} MB')
        
    # ~ saving metadata
    # Alongside the actual dataset information, we also save metadata about the dataset which will later 
    # on be available to be fetched from the remote file share server and includes information such as 
    # a short description about the dataset, some relevant tags but also automatically determined information 
    # such as the number of elements in the dataset and the number of 
    e.log('saving metadata...')
        
    # first of all we need to construct the actual metadata dict.
    example_graph = graphs[0]
    pprint(example_graph)
    metadata: dict = {
        'compounds': len(graphs),
        'targets': len(example_graph['graph_labels']),
        'target_type': [e.DATASET_TYPE],
        'description': e.DESCRIPTION,
        'raw': ['csv'],
        'sources': [],
    }
    
    # We also want to apply the user-defined overwrites to the metadata dict.
    metadata.update(e.METADATA)
    
    # Finally we save the metadata in yml format to the experiment archive folder.
    metadata_path = os.path.join(e.path, 'metadata.yml')
    with open(metadata_path, 'w') as metadata_file:
        yaml.dump(metadata, metadata_file)
        
    e.log(f'saved metadata @ {metadata_path}')
        
    # :hook after_save:
    #       Action hook that is called after the graph dataset itself has been saved to the disk inside 
    #       the experiment archive folder. Receives the dataset in the form of the index_data_map and the 
    #       list of graphs as parameters.
    e.apply_hook(
        'after_save',
        index_data_map=dataset,
        graphs=graphs,
    )

experiment.run_if_main()


"""
In this file we simply test if importing from the main package name works correctly.
"""


def test_import_ensure_dataset():
    """
    It should be possible to import the ``ensure_dataset`` function directly using the 
    top-level package name.
    """
    from chem_mat_data import ensure_dataset
    assert callable(ensure_dataset)


from collections import defaultdict

import numpy as np

import chem_mat_data._typing as tc
from typing import Dict, List
from scipy.spatial.distance import pdist, squareform


# == UTILITY ==


def to_graph_dict(data: dict) -> tc.GraphDict:
    """
    Ensures that the given dictionary ``data`` passes as a graph dict representation 
    in the sense that this function will convert all the list properties of the given dict 
    into numpy arrays - as it is required by the GraphDict format.
    
    :param data: The dictionary to be converted.
    
    :returns: The orgininal dict, but with all list elements converted into numpy arrays
    """
    for key, value in data.items():
        if isinstance(value, list):
            data[key] = np.array(value)
    
    return data
    
    
def graph_adjacency_list(graph: tc.GraphDict) -> Dict[int, List]:
    """
    Creates an adjacency list / dict from the given GraphDict ``graph``. Returns a dictionary 
    where the keys are the node indices of the given graph and the values are again lists of 
    node indices - containing exactly those node indices which the corresponding node is 
    directly connected to.
    
    :param graph: The graph for which the adjacency list should be created.
    
    :returns: The dict structure which maps "node index -> list of connected node indices"
    """
    adjacency_list: dict = defaultdict(set)
    for i, j in graph['edge_indices']:
        adjacency_list[i].append(j)
    
    return adjacency_list


def _visitation_dfs(node: int, 
                    adjacency_list: dict, 
                    visited: np.ndarray
                    ) -> None:
    """
    Helper function which performs depth first search for the given node and adjacency list.
    """
    visited[node] = 1
    for neighbor in adjacency_list[node]:
        if not visited[neighbor]:
            _visitation_dfs(neighbor, adjacency_list, visited)

    
def is_graph_connected(graph: tc.GraphDict) -> None:
    """
    This function returns the boolean property of whether the ``graph`` is connected or not. 
    This means that from every node of the graph it should be possible to reach every 
    other node of the graph in a finite number of steps.
    
    :param graph: The GraphDict to be checked.
    
    :returns: boolean value
    """
    adjacency_list = graph_adjacency_list(graph)
    visited = np.zeros_like(graph['node_indices'])
    _visitation_dfs(0, adjacency_list, visited)
    
    return (visited > 0).all()


def graph_threshold_edges(graph: tc.GraphDict, 
                          threshold: float,
                          ) -> tc.GraphDict:
    """
    Given a ``graph`` dict representation and a distance ``threshold`` this function will calculate all the 
    pairwise distances between the nodes of the graph and then create new edges by inserting new edges for all 
    nodes whose distance is smaller or equal to the given threshold. The function will also replace the previous 
    edge_attributes of the 
    """
    assert 'node_coordinates' in graph, (
        'graph must contain "node_coordinates" property to calculate distance-based edges. '
    )
    
    node_indices: np.ndarray = graph['node_indices']
    node_coordinates: np.ndarray = graph['node_coordinates']
    # Calculate pairwise distances between all nodes
    distances = squareform(pdist(node_coordinates))

    # Create new edge list based on the threshold
    new_edge_indices = []
    for i in range(len(node_indices)):
        for j in range(i + 1, len(node_indices)):
            if distances[i, j] <= threshold:
                new_edge_indices.append([i, j])

    graph['edge_indices'] = np.array(new_edge_indices, dtye=np.int32)
    graph['edge_attributes'] = np.ones((len(new_edge_indices), 1), dtype=np.float32)
    return graph

# == ASSERTIONS ==

def assert_graph_dict(graph: tc.GraphDict) -> None:
    """
    Wraps a series of assert statements which make sure that the given ``graph`` object is a 
    valid GraphDict representation.
    
    This will for example make sure that the dict contains the required properties, that the 
    values of those properties are proper numpy arrays and that those arrays have the 
    correct shape.
    """
    # The most basic requirement is that the given object is in fact a non-empty dictionary 
    assert isinstance(graph, dict), 'graph dict must be dictionary'
    assert len(graph) != 0, 'graph dict must not be empty'
    
    # Even though the GraphDict representation is somewhat flexible there are some attributes 
    # which are absolutely required and which we will test for here.
    # Additionally, all of those properties need to be numpy arrays or need to be convertible 
    # into numpy arrays.
    required_keys = [
        'node_indices',
        'node_attributes',
        'edge_indices',
        'edge_attributes',
    ]
    for key in required_keys:
        assert key in graph, f'graph dict must contain "{key}" property'
        assert isinstance(graph[key], np.ndarray), f'graph property "{key}" needs to be numpy array'
    
    # ~ specific checks
    
    # The number of nodes needs to be conistent for all node properties
    num_nodes = len(graph['node_indices'])
    for key, value in graph.items():
        if key.startswith('node'):
            assert isinstance(value, np.ndarray), f'node property "{key}" needs to be numpy array'
            assert value.shape[0] == num_nodes, f'node property "{key}" needs to match no nodes {num_nodes}'
    
    # The number of edges needs to be consistent for all edge properties
    num_edges = len(graph['edge_indices'])
    for key, value in graph.items():
        if key.startswith('edge'):
            assert isinstance(value, np.ndarray), f'edge property "{key}" needs to be numpy array'
            assert value.shape[0] == num_edges, f'edge property "{key}" needs to match no edges {num_edges}'
    



import os
import tempfile

from chem_mat_data.cache import Cache
from chem_mat_data.config import Config


class TestConfig:
    """
    Unit tests for the Config class.
    """
    def test_construction_basically_works(self):
        """
        It should be possible to construct the config singleton and it should internally link to the config data dict 
        that was read from the config file and it should also link to the Cache wrapper object that manages the cache 
        folder.
        """
        config = Config()
        assert isinstance(config, Config)
        
        # This should be a singleton instance, which means that no matter how many times we create a new
        # instance of the Config class, it should always be the same object.
        config2 = Config()
        assert config is config2
        
        # Some config toml file should have been created
        assert os.path.exists(config.config_file_path)
        assert config.config_file_path.endswith('.toml')
        
        # The default version of the config data should be loaded from the config toml file
        assert isinstance(config.config_data, dict)
        assert len(config.config_data) != 0
        
        # Additionally, the cache wrapper object should be linked in the config object and 
        # it should point to a cache folder that exists
        assert isinstance(config.cache, Cache)
        assert os.path.exists(config.cache.path)
        
    def test_set_cache_path_works(self):
        """
        It should be possible to switch the cache folder by using the "set_cache_path" method on the config
        object.
        """
        with tempfile.TemporaryDirectory() as path:
            
            config = Config()
            # Before changing the path the cache folder should not be the new path
            prev_path = config.cache_path
            assert config.cache_path != path
            assert config.cache.path != path
            
            # Only after changing, the cache folder should be the new path. It is also important that 
            # the Cache wrapper object also points to this new path!
            config.set_cache_path(path)
            assert config.cache_path == path
            assert config.cache.path == path
            
            # After the test we need to reset the cache path to the previous path to avoid potential side effects
            config.set_cache_path(prev_path)
            
    def test_set_config_path_works(self):
        """
        It should be possible to switch the config folder by using the "set_config_path" method on the config
        object.
        """
        with tempfile.TemporaryDirectory() as path:
            
            config = Config()
            prev_path = config.config_path
            assert config.config_path != path
            
            config.set_config_path(path)
            assert config.config_path == path
            assert os.path.exists(config.config_file_path)
            # Since this operation should have created a new config file in the tempdir, this folder should now 
            # no longer be empty.
            assert len(os.listdir(path)) != 0
            
            # After the test we need to reset the config path to the previous path to avoid potential side effects
            config.set_config_path(prev_path)


"""
This experiment creates the processed graph dataset for **tmQMg** — a dataset of ~63,000
mononuclear transition metal complexes with 20 DFT-computed quantum-mechanical properties.

The raw data is downloaded from the ELECTRUM validation repository on GitHub. Each complex
is represented in a decomposed format: a metal element symbol and dot-separated ligand
SMILES strings. The script infers connecting atom indices heuristically (identifying
donor atoms with available lone pairs) and uses ``MetalOrganicProcessing`` to build the
molecular graphs.

**Source**: Kneiding et al., *Digital Discovery* 2023; ELECTRUM repo (Orsi & Frei 2025).

**Targets** (20 regression properties):
    HOMO/LUMO energies, HOMO-LUMO gap, electronic energy, dispersion energy, enthalpy,
    Gibbs energy, ZPE correction, heat capacity, entropy, dipole moment, polarizability,
    vibrational frequencies, and delta corrections.

**Usage**::

    python create_graph_datasets__tmqmg.py

Results are written to ``results/create_graph_datasets__tmqmg/debug/``.
"""
import os
import gzip
import time
import shutil
import datetime
import tempfile
import multiprocessing
from typing import Dict, List, Optional

import yaml
import msgpack
import requests
import numpy as np
import rdkit.Chem as Chem
import pandas as pd
from rich.pretty import pprint
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

from rdkit import RDLogger
RDLogger.DisableLog('rdApp.*')

import chem_mat_data._typing as typc
from chem_mat_data.tmc_processing import MetalOrganicProcessing, TRANSITION_METAL_ATOMIC_NUMBERS
from chem_mat_data.data import default, ext_hook, save_graphs


# == SOURCE PARAMETERS ==
# These parameters configure where the raw dataset is fetched from and how it is
# interpreted. The tmQMg dataset is hosted in the ELECTRUM validation repository
# on GitHub as a single CSV file with columns for CSD identifier, metal, charge,
# ligand SMILES (dot-separated), and 20 DFT-computed QM properties.

# :param DOWNLOAD_URL:
#       The URL from which the raw tmQMg CSV file will be downloaded. This points
#       to the ELECTRUM validation repository's ``datasets/tmQMg.csv`` on GitHub.
#       The file is approximately 20 MB and contains ~63,000 rows.
DOWNLOAD_URL: str = 'https://raw.githubusercontent.com/TheFreiLab/electrum_val/main/datasets/tmQMg.csv'

# :param TARGET_COLUMNS:
#       A list of the 20 column names in the raw CSV that contain the DFT-computed
#       quantum-mechanical properties. These become the regression targets (graph_labels)
#       in the processed graph dataset. The order here determines the order in the
#       target vector.
TARGET_COLUMNS: List[str] = [
    'tzvp_lumo_energy',
    'tzvp_homo_energy',
    'tzvp_homo_lumo_gap',
    'homo_lumo_gap_delta',
    'tzvp_electronic_energy',
    'electronic_energy_delta',
    'tzvp_dispersion_energy',
    'dispersion_energy_delta',
    'enthalpy_energy',
    'enthalpy_energy_correction',
    'gibbs_energy',
    'gibbs_energy_correction',
    'zpe_correction',
    'heat_capacity',
    'entropy',
    'tzvp_dipole_moment',
    'dipole_moment_delta',
    'polarisability',
    'lowest_vibrational_frequency',
    'highest_vibrational_frequency',
]

# :param DATASET_TYPE:
#       Either 'regression' or 'classification'. All 20 tmQMg targets are continuous
#       QM properties, so this is always 'regression'.
DATASET_TYPE: str = 'regression'

# :param DESCRIPTION:
#       A human-readable description of the dataset that will be stored in the
#       metadata.yml file alongside the processed dataset.
DESCRIPTION: str = (
    'Graph dataset of ~63,000 mononuclear transition metal complexes from the tmQMg dataset '
    '(Kneiding et al., Digital Discovery 2023) with 20 DFT-computed quantum-mechanical '
    'properties. Processed using MetalOrganicProcessing with decomposed metal + ligand SMILES '
    'representation and heuristic donor atom detection.'
)

# :param METADATA:
#       A dictionary of additional metadata fields that will be merged into the
#       auto-generated metadata.yml. Includes tags for discoverability, source
#       URLs for provenance, and per-target descriptions.
METADATA: dict = {
    'category': 'tmc',
    'min_version': '1.7.0',
    'tags': ['Molecules', 'TransitionMetals', 'TMC', 'QM', 'DFT'],
    'sources': [
        'https://doi.org/10.1039/D2DD00129B',
        'https://github.com/TheFreiLab/electrum_val',
    ],
    'verbose': 'tmQMg Transition Metal Complex QM Properties',
    'target_descriptions': {
        str(i): col for i, col in enumerate(TARGET_COLUMNS)
    },
}

# == PROCESSING PARAMETERS ==
# These parameters control how the dataset is processed and stored. Unlike the
# organic molecule pipeline which uses MoleculeProcessing, this script uses
# MetalOrganicProcessing which takes a decomposed input (metal + ligand SMILES)
# rather than a single whole-complex SMILES string.

# :param DATASET_NAME:
#       The base filename for all output files (CSV, mpack, metadata). The script
#       produces: ``{DATASET_NAME}.csv``, ``{DATASET_NAME}.mpack``,
#       ``{DATASET_NAME}.mpack.gz``, and ``metadata.yml``.
DATASET_NAME: str = 'tmqmg'

# :param COMPRESS:
#       If True, the mpack file is additionally compressed to gzip format. This
#       typically achieves 10-15x compression and is recommended for distribution.
COMPRESS: bool = True

# :param MAX_ELEMENTS:
#       Maximum number of complexes to process. Set to None to process the full
#       dataset (~63,000 complexes). Set to a smaller number (e.g. 100) for quick
#       verification runs during development.
MAX_ELEMENTS: Optional[int] = None

# == EXPERIMENT PARAMETERS ==

# :param __DEBUG__:
#       In debug mode, results overwrite the previous run in the ``debug/`` folder
#       instead of creating a new timestamped folder.
__DEBUG__ = True

# :param __TESTING__:
#       If True, the dataset is limited to 50 elements for fast testing regardless
#       of MAX_ELEMENTS.
__TESTING__ = False

experiment = Experiment(
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)


# ======================================================================================
# DONOR ATOM HEURISTIC
# ======================================================================================
# The tmQMg dataset provides ligand SMILES but does NOT provide information about which
# atoms in each ligand coordinate to the metal. We need to infer this heuristically.
#
# The approach: in coordination chemistry, the atoms that donate electron pairs to the
# metal are almost always heteroatoms (N, O, P, S, Se) or halides (F, Cl, Br, I), and
# anionic donors (negatively charged atoms) take priority. This covers the vast majority
# of ligands in the dataset — amines, phosphines, carbonyls, halides, carboxylates, etc.
#
# The main limitation is carbon-donor ligands (NHC carbenes, cyclopentadienyl, CO) where
# carbon is the actual donor atom. For these, we fall back to index 0 which happens to be
# correct for CO (C#O, carbon is index 0) but may be wrong for more complex cases.

# Atomic numbers of common donor atoms in coordination chemistry.
DONOR_ELEMENTS: set = {
    7,   # N
    8,   # O
    15,  # P
    16,  # S
    34,  # Se
    9,   # F
    17,  # Cl
    35,  # Br
    53,  # I
}


def infer_connecting_atoms(ligand_smiles: str) -> List[int]:
    """
    Heuristically determine which atoms in a ligand SMILES are donor atoms
    (i.e., atoms that coordinate to the metal center).

    The heuristic identifies atoms that:

    1. Are common donor elements (N, O, P, S, Se, halides), OR
    2. Carry a negative formal charge (anionic donors like carbanions).

    Among candidate donors, atoms with negative charge are prioritized. For neutral
    ligands where multiple candidates exist, all candidates are returned (the ligand
    may be polydentate).

    For single-atom ligands (e.g., ``Cl``, ``Br``), atom index 0 is always returned.

    :param ligand_smiles: SMILES string of a single ligand fragment.
    :returns: List of 0-based atom indices of inferred donor atoms.
    """
    mol = Chem.MolFromSmiles(ligand_smiles)
    if mol is None:
        # Some ligand fragments have unusual valences (e.g., B with 4 bonds in
        # tris(pyrazolyl)borate) because they were extracted from a metal complex.
        # Fall back to partial sanitization that skips the valence check.
        mol = Chem.MolFromSmiles(ligand_smiles, sanitize=False)
        if mol is not None:
            try:
                Chem.SanitizeMol(mol, Chem.SanitizeFlags.SANITIZE_ALL ^ Chem.SanitizeFlags.SANITIZE_PROPERTIES)
            except Exception:
                pass
    if mol is None:
        return [0]

    num_atoms = mol.GetNumAtoms()
    if num_atoms == 1:
        return [0]

    # Collect candidates in two tiers: negatively charged atoms (strongest donors)
    # and neutral heteroatoms (common donors).
    candidates = []
    charged_candidates = []
    for atom in mol.GetAtoms():
        z = atom.GetAtomicNum()
        charge = atom.GetFormalCharge()

        if charge < 0:
            charged_candidates.append(atom.GetIdx())
        elif z in DONOR_ELEMENTS:
            candidates.append(atom.GetIdx())

    # Prefer negatively charged atoms (anionic donors like carboxylate O-)
    if charged_candidates:
        return charged_candidates

    # Fall back to neutral donor atoms (amines, phosphines, etc.)
    if candidates:
        return candidates

    # Last resort for carbon-donor ligands (CO, isocyanides, NHC carbenes):
    # return index 0. For CO (C#O) this is correct since C is the donor.
    return [0]


# ======================================================================================
# PROCESSING WORKER
# ======================================================================================
# This worker mirrors the ProcessingWorker in the base ``create_graph_datasets.py`` but
# uses ``MetalOrganicProcessing`` instead of ``MoleculeProcessing``. The key difference
# is the input format: instead of a single SMILES string, each work item contains the
# decomposed representation (metal, ligand_smiles list, connecting_atom_indices).

class TMCProcessingWorker(multiprocessing.Process):
    """
    Multiprocessing worker that converts decomposed TMC data into graph representations
    using ``MetalOrganicProcessing``.

    Communication with the main process happens via two queues: the ``input_queue``
    delivers data dicts to be processed and the ``output_queue`` returns msgpack-encoded
    graph dicts (or None on failure) back to the main process.
    """

    def __init__(self,
                 input_queue: multiprocessing.Queue,
                 output_queue: multiprocessing.Queue,
                 ):
        super().__init__()
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.processing = MetalOrganicProcessing()

    def run(self):
        """
        Infinite loop consuming data dicts from the input queue. For each item,
        calls ``MetalOrganicProcessing.process()`` with the decomposed TMC fields
        and sends the resulting graph (or None on error) to the output queue.
        A ``None`` sentinel on the input queue signals the worker to stop.
        """
        for data in iter(self.input_queue.get, None):
            try:
                graph: typc.GraphDict = self.processing.process(
                    metal=data['metal'],
                    ligand_smiles=data['ligand_smiles'],
                    connecting_atom_indices=data['connecting_atom_indices'],
                    oxidation_state=data.get('oxidation_state', 0),
                    total_charge=data.get('total_charge', 0),
                    spin_multiplicity=data.get('spin_multiplicity', 1),
                    whole_complex_smiles=data.get('whole_complex_smiles', None),
                )
                graph['graph_labels'] = np.array(data['targets'])

                experiment.apply_hook(
                    'add_graph_metadata',
                    data=data,
                    graph=graph,
                )

            except Exception as exc:
                metal = data.get('metal', '?')
                cid = data.get('complex_id', '?')
                print(f' ! error processing {cid} ({metal}) - {exc.__class__.__name__}: {exc}')
                graph = None

            graph_encoded = msgpack.packb(graph, default=default)
            self.output_queue.put(graph_encoded)


# ======================================================================================
# HOOKS
# ======================================================================================

@experiment.hook('add_graph_metadata', default=True, replace=False)
def add_graph_metadata(e: Experiment, data: dict, graph: typc.GraphDict) -> None:
    """
    This hook is invoked in the processing worker after the graph has been created.
    It adds the CSD identifier and metal symbol to each graph dict for traceability,
    so that processed graphs can be linked back to their source entries.
    """
    if graph is not None:
        graph['graph_id'] = data.get('complex_id', '')
        graph['graph_metal'] = data.get('metal', '')


@experiment.hook('load_dataset', default=True, replace=False)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    """
    Download the tmQMg CSV from the ELECTRUM repo and parse it into the decomposed
    TMC format expected by ``MetalOrganicProcessing``.

    The raw CSV has a ``LigandSmiles`` column that contains all ligands for a complex
    concatenated with dots (the standard SMILES disconnection notation). For example,
    ``"C#O.C#O.Cl.Cl"`` represents two carbonyl and two chloride ligands. This hook
    splits them into individual ligand SMILES and infers which atoms in each ligand
    coordinate to the metal using the ``infer_connecting_atoms`` heuristic.

    Note that tmQMg does not provide oxidation states or spin multiplicities, so these
    default to 0 and 1 respectively.
    """
    # -- Download the raw CSV (cached after the first run) --
    csv_path = os.path.join(e.path, 'tmqmg_raw.csv')
    if not os.path.exists(csv_path):
        e.log(f'downloading tmQMg dataset from {e.DOWNLOAD_URL} ...')
        response = requests.get(e.DOWNLOAD_URL, timeout=120)
        response.raise_for_status()
        with open(csv_path, 'w') as f:
            f.write(response.text)
        e.log(f'downloaded {len(response.content) / 1024 / 1024:.1f} MB')
    else:
        e.log(f'using cached raw CSV at {csv_path}')

    # -- Parse each row into the decomposed TMC format --
    e.log('parsing dataset...')
    df = pd.read_csv(csv_path)

    if e.MAX_ELEMENTS is not None:
        df = df.head(e.MAX_ELEMENTS)
        e.log(f'limited to {e.MAX_ELEMENTS} complexes')

    dataset: Dict[int, dict] = {}
    skipped = 0

    for idx, row in df.iterrows():
        metal = row['Metal']
        ligand_smiles_raw = str(row['LigandSmiles'])
        charge = int(row['charge'])

        # The LigandSmiles column uses the SMILES dot notation to separate individual
        # ligand fragments. Each fragment between dots is one complete ligand structure.
        ligand_smiles = ligand_smiles_raw.split('.')

        # For each ligand fragment, heuristically determine which atoms are donor atoms
        # that coordinate to the metal center. This is necessary because tmQMg does not
        # provide explicit connecting atom information.
        connecting_atom_indices = []
        valid = True
        for lig_smi in ligand_smiles:
            conn = infer_connecting_atoms(lig_smi)
            if not conn:
                valid = False
                break
            connecting_atom_indices.append(conn)

        if not valid:
            skipped += 1
            continue

        # Collect the 20 QM property values as the regression target vector
        targets = []
        for col in e.TARGET_COLUMNS:
            targets.append(float(row[col]))

        dataset[len(dataset)] = {
            'complex_id': str(row['id']),
            'metal': metal,
            'ligand_smiles': ligand_smiles,
            'connecting_atom_indices': connecting_atom_indices,
            'total_charge': charge,
            'oxidation_state': 0,      # Not available in tmQMg
            'spin_multiplicity': 1,    # Not available in tmQMg
            'targets': targets,
            # Keep the original dot-separated string for CSV output
            'ligand_smiles_raw': ligand_smiles_raw,
        }

    e.log(f'parsed {len(dataset)} complexes ({skipped} skipped)')
    return dataset


@experiment.hook('save_csv', default=True, replace=False)
def save_csv(e: Experiment, dataset: Dict[int, dict]) -> None:
    """
    Save the dataset as a CSV file in the decomposed TMC format. This CSV serves as
    the "raw" format counterpart to the processed mpack graph file and can be loaded
    via ``load_tmc_dataset('tmqmg')``.
    """
    e.log('saving CSV...')

    rows = []
    for index, data in dataset.items():
        row = {
            'complex_id': data['complex_id'],
            'metal': data['metal'],
            'ligand_smiles': data['ligand_smiles_raw'],
            'total_charge': data['total_charge'],
            'oxidation_state': data['oxidation_state'],
            'spin_multiplicity': data['spin_multiplicity'],
        }
        for i, col in enumerate(e.TARGET_COLUMNS):
            row[col] = data['targets'][i]
        rows.append(row)

    df = pd.DataFrame(rows)

    csv_path = os.path.join(e.path, f'{e.DATASET_NAME}.csv')
    df.to_csv(csv_path, index=False)

    # Also save a compressed version for distribution
    gz_path = csv_path + '.gz'
    with open(csv_path, 'rb') as f_in, gzip.open(gz_path, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)

    e.log(f'saved CSV ({len(rows)} rows)')


# ======================================================================================
# MAIN EXPERIMENT
# ======================================================================================

@experiment
def experiment(e: Experiment):

    e.log('starting tmQMg processing experiment...')
    e.log_parameters()

    # -- Create worker processes --
    # Workers are spawned before loading the dataset to avoid copying the full dataset
    # into each subprocess's memory (Python forks the current process, including its
    # memory, when spawning subprocesses).
    e.log('creating TMC processing workers...')
    input_queue = multiprocessing.Queue(maxsize=100)
    output_queue = multiprocessing.Queue(maxsize=100)
    workers = []
    num_workers = os.cpu_count()
    for _ in range(num_workers):
        worker = TMCProcessingWorker(
            input_queue=input_queue,
            output_queue=output_queue,
        )
        worker.start()
        workers.append(worker)

    e.log(f'started {num_workers} workers')

    try:
        # -- Load and parse the dataset --
        e.log('loading dataset...')
        dataset: Dict[int, dict] = e.apply_hook('load_dataset')
        num_elements = len(dataset)
        e.log(f'loaded {num_elements} complexes')

        # -- Save the raw CSV alongside the processed data --
        e.apply_hook('save_csv', dataset=dataset)

        # In testing mode, limit to a small subset so the full pipeline can be
        # validated quickly without processing the entire dataset.
        if e.__TESTING__:
            e.log('testing mode: limiting to 50 elements')
            num_elements = min(num_elements, 50)
            dataset = dict(list(dataset.items())[:num_elements])

        # -- Feed data to workers and collect results --
        # The main process continuously fills the input queue and drains the output
        # queue in a loop. Workers convert each decomposed TMC entry into a GraphDict
        # using MetalOrganicProcessing. Failed conversions produce None and are silently
        # dropped from the final dataset (with an error message printed to stdout).
        e.log('processing dataset...')
        indices = list(dataset.keys())
        num_indices = len(indices)
        graphs = []

        start_time = time.time()
        count = 0
        prev_count = 0

        while count < num_indices:

            # Keep the input queue full so workers always have work available
            while not input_queue.full() and len(indices) != 0:
                index = indices.pop()
                data = dataset[index]
                input_queue.put(data)

            # Drain completed results from the output queue
            while not output_queue.empty():
                graph_encoded = output_queue.get()
                graph = msgpack.unpackb(graph_encoded, ext_hook=ext_hook)
                if graph:
                    graphs.append(graph)
                count += 1

            # Progress logging every 1000 elements with ETA estimate
            if count % 1000 == 0 and count != prev_count:
                prev_count = count
                time_passed = time.time() - start_time
                num_remaining = num_elements - (count + 1)
                time_per_element = time_passed / (count + 1)
                time_remaining = time_per_element * num_remaining
                eta = datetime.datetime.now() + datetime.timedelta(seconds=time_remaining)
                e.log(f' * {count:05d}/{num_elements} done'
                      f' - time passed: {time_passed / 60:.2f}m'
                      f' - eta: {eta:%a %d.%m %H:%M}')

        end_time = time.time()
        duration = end_time - start_time
        e.log(f'finished processing {len(graphs)} graphs in {duration:.1f}s')

    finally:
        # Ensure workers are stopped even if an exception occurs. Each worker
        # terminates when it receives a None sentinel on its input queue.
        e.log('stopping workers...')
        for worker in workers:
            input_queue.put(None)
            worker.terminate()
            worker.join()

        del input_queue
        del output_queue

    # -- Save processed graphs as msgpack --
    e.log(f'saving {len(graphs)} graphs to mpack...')
    dataset_path = os.path.join(e.path, e.DATASET_NAME + '.mpack')
    save_graphs(graphs, dataset_path)

    file_size = os.path.getsize(dataset_path)
    e.log(f'wrote mpack: {file_size / 1024 / 1024:.1f} MB')

    # -- Compress the mpack file for distribution --
    if e.COMPRESS:
        e.log('compressing...')
        compressed_path = os.path.join(e.path, e.DATASET_NAME + '.mpack.gz')
        with open(dataset_path, 'rb') as f_in, gzip.open(compressed_path, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)

        compressed_size = os.path.getsize(compressed_path)
        e.log(f'compressed: {compressed_size / 1024 / 1024:.1f} MB')

    # -- Generate and save metadata --
    # The metadata file combines auto-detected information (compound count, target count)
    # with the user-defined METADATA dict (tags, sources, descriptions).
    e.log('saving metadata...')
    example_graph = graphs[0]
    pprint(example_graph)

    metadata: dict = {
        'compounds': len(graphs),
        'targets': len(example_graph['graph_labels']),
        'target_type': [e.DATASET_TYPE],
        'description': e.DESCRIPTION,
        'raw': ['csv'],
        'sources': [],
    }
    metadata.update(e.METADATA)

    metadata_path = os.path.join(e.path, 'metadata.yml')
    with open(metadata_path, 'w') as f:
        yaml.dump(metadata, f)

    e.log(f'saved metadata @ {metadata_path}')
    e.log('done!')


experiment.run_if_main()
